#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import os
import pytest
import ezdxf
from ezdxf.lldxf.tagger import ascii_tags_loader
from ezdxf.lldxf.loader import load_dxf_structure

BASEDIR = os.path.dirname(__file__)
DATADIR = "data"


@pytest.fixture(params=["POLI-ALL210_12.DXF"])
def filename(request):
    filename = os.path.join(BASEDIR, DATADIR, request.param)
    if not os.path.exists(filename):
        pytest.skip(f"File {filename} not found.")
    return filename


def test_check_R12_has_handles(filename):
    dwg = ezdxf.readfile(filename)
    assert dwg.header["$HANDLING"] > 0
    for entity in dwg.modelspace():
        assert int(entity.dxf.handle, 16) > 0


def test_write_R12_without_handles(filename, tmpdir):
    dwg = ezdxf.readfile(filename)
    dwg.header["$HANDLING"] = 0
    export_path = str(tmpdir.join("dxf_r12_without_handles.dxf"))
    dwg.saveas(export_path)

    # can't check with ezdxf.readfile(), because handles automatically enabled
    with open(export_path) as f:
        tagger = ascii_tags_loader(f)
        sections = load_dxf_structure(tagger)
        for entity in sections["ENTITIES"]:
            with pytest.raises(ezdxf.DXFValueError):  # has no handles
                entity.get_handle()

        for entity in sections["TABLES"]:
            if entity[0] != (0, "DIMSTYLE"):
                with pytest.raises(ezdxf.DXFValueError):  # has no handles
                    entity.get_handle()
            else:  # special DIMSTYLE entity
                assert (
                    len(entity.find_all(105)) == 0
                ), "remove handle group code 105"
                assert (
                    len(entity.find_all(5)) == 1
                ), "do not remove group code 5"



# Copyright (c) 2011-2021 Manfred Moitzi
# License: MIT License

from . import factory

# basic classes
from .xdict import ExtensionDict
from .xdata import XData
from .appdata import AppData, Reactors
from .dxfentity import DXFEntity, DXFTagStorage
from .dxfgfx import DXFGraphic, SeqEnd, is_graphic_entity, get_font_name
from .dxfobj import DXFObject, is_dxf_object
from .dxfns import DXFNamespace, SubclassProcessor

# register management structures
from .dxfclass import DXFClass
from .table import TableHead

# register table entries
from .ltype import Linetype
from .layer import Layer, LayerOverrides
from .textstyle import Textstyle
from .dimstyle import DimStyle
from .view import View
from .vport import VPort
from .ucs import UCSTableEntry
from .appid import AppID
from .blockrecord import BlockRecord

# register DXF objects R2000
from .acad_proxy_entity import ACADProxyEntity
from .dxfobj import XRecord, Placeholder, VBAProject, SortEntsTable
from .dictionary import Dictionary, DictionaryVar, DictionaryWithDefault
from .layout import DXFLayout
from .idbuffer import IDBuffer
from .sun import Sun
from .material import Material, MaterialCollection
from .oleframe import OLE2Frame
from .spatial_filter import SpatialFilter

# register DXF objects R2007
from .visualstyle import VisualStyle

# register entities R12
from .line import Line
from .point import Point
from .circle import Circle
from .arc import Arc
from .shape import Shape
from .solid import Solid, Face3d, Trace
from .text import Text
from .subentity import LinkedEntities, entity_linker
from .insert import Insert
from .block import Block, EndBlk
from .polyline import Polyline, Polyface, Polymesh, MeshVertexCache
from .attrib import Attrib, AttDef, copy_attrib_as_text
from .dimension import *
from .dimstyleoverride import DimStyleOverride
from .viewport import Viewport

# register graphical entities R2000
from .lwpolyline import LWPolyline
from .ellipse import Ellipse
from .xline import XLine, Ray
from .mtext import MText
from .mtext_columns import *
from .spline import Spline
from .mesh import Mesh, MeshData
from .boundary_paths import *
from .gradient import *
from .pattern import *
from .hatch import *
from .mpolygon import MPolygon
from .image import Image, ImageDef, Wipeout
from .underlay import (
    Underlay,
    UnderlayDefinition,
    PdfUnderlay,
    DgnUnderlay,
    DwfUnderlay,
)
from .leader import Leader
from .tolerance import Tolerance
from .helix import Helix
from .acis import (
    Body,
    Solid3d,
    Region,
    Surface,
    ExtrudedSurface,
    LoftedSurface,
    RevolvedSurface,
    SweptSurface,
)
from .mline import MLine, MLineVertex, MLineStyle, MLineStyleCollection
from .mleader import MLeader, MLeaderStyle, MLeaderStyleCollection, MultiLeader

# register graphical entities R2004

# register graphical entities R2007

from .light import Light
from .acad_table import (
    AcadTableBlockContent,
    acad_table_to_block,
    read_acad_table_content,
)

# register graphical entities R2010

from .geodata import GeoData

# register graphical entities R2013

# register graphical entities R2018



# Copyright (c) 2020-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Iterable,
    Tuple,
    Union,
    cast,
    Sequence,
    Optional,
)
from typing_extensions import TypeAlias
from abc import abstractmethod
from collections import namedtuple
import math
import numpy as np

from ezdxf.lldxf.const import VTX_EXTRA_VERTEX_CREATED, VTX_SPLINE_FRAME_CONTROL_POINT
from ezdxf.math import (
    Vec2,
    Vec3,
    UVec,
    BSpline,
    ConstructionRay,
    OCS,
    ParallelRaysError,
    bulge_to_arc,
    ConstructionArc,
)

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFGraphic, Solid, Trace, Face3d, LWPolyline, Polyline

__all__ = ["TraceBuilder", "LinearTrace", "CurvedTrace"]

LinearStation = namedtuple("LinearStation", ("vertex", "start_width", "end_width"))
# start_width of the next (following) segment
# end_width of the next (following) segment

CurveStation = namedtuple("CurveStation", ("vertex0", "vertex1"))

Face: TypeAlias = Tuple[Vec2, Vec2, Vec2, Vec2]
Polygon: TypeAlias = Sequence[Vec2]
Quadrilateral: TypeAlias = Union["Solid", "Trace", "Face3d"]


class AbstractTrace:
    @abstractmethod
    def faces(self) -> Iterable[Face]:
        # vertex order: up1, down1, down2, up2
        # faces connections:
        # up2 -> next up1
        # down2 -> next down1
        pass

    def polygon(self) -> Polygon:
        def merge(vertices: Polygon) -> Iterable[UVec]:
            if not len(vertices):
                return

            _vertices = iter(vertices)
            prev = next(_vertices)
            yield prev
            for vertex in _vertices:
                if not prev.isclose(vertex):
                    yield vertex
                    prev = vertex

        forward_contour: list[Vec2] = []
        backward_contour: list[Vec2] = []
        for up1, down1, down2, up2 in self.faces():
            forward_contour.extend((down1, down2))
            backward_contour.extend((up1, up2))

        contour = list(merge(forward_contour))
        contour.extend(reversed(list(merge(backward_contour))))
        return contour

    def virtual_entities(
        self, dxftype="TRACE", dxfattribs=None, doc: Optional[Drawing] = None
    ) -> Iterable[Quadrilateral]:
        """
        Yields faces as SOLID, TRACE or 3DFACE entities with DXF attributes
        given in `dxfattribs`.

        If a document is given, the doc attribute of the new entities will be
        set and the new entities will be automatically added to the entity
        database of that document.

        Args:
            dxftype: DXF type as string, "SOLID", "TRACE" or "3DFACE"
            dxfattribs: DXF attributes for SOLID, TRACE or 3DFACE entities
            doc: associated document

        """
        from ezdxf.entities.factory import new

        if dxftype not in {"SOLID", "TRACE", "3DFACE"}:
            raise TypeError(f"Invalid dxftype {dxftype}.")
        dxfattribs = dict(dxfattribs or {})
        for face in self.faces():
            for i in range(4):
                dxfattribs[f"vtx{i}"] = face[i]

            if dxftype != "3DFACE":
                # weird vertex order for SOLID and TRACE
                dxfattribs["vtx2"] = face[3]
                dxfattribs["vtx3"] = face[2]
            entity = new(dxftype, dxfattribs, doc)
            if doc:
                doc.entitydb.add(entity)
            yield entity  # type: ignore


class LinearTrace(AbstractTrace):
    """Linear 2D banded lines like polylines with start- and end width.

    Accepts 3D input, but z-axis is ignored.

    """

    def __init__(self) -> None:
        self._stations: list[LinearStation] = []
        self.abs_tol = 1e-12

    def __len__(self):
        return len(self._stations)

    def __getitem__(self, item):
        return self._stations[item]

    @property
    def is_started(self) -> bool:
        """`True` if at least one station exist."""
        return bool(self._stations)

    def add_station(
        self, point: UVec, start_width: float, end_width: Optional[float] = None
    ) -> None:
        """Add a trace station (like a vertex) at location `point`,
        `start_width` is the width of the next segment starting at this station,
        `end_width` is the end width of the next segment.

        Adding the last location again, replaces the actual last location e.g.
        adding lines (a, b), (b, c), creates only 3 stations (a, b, c), this is
        very important to connect to/from splines.

        Args:
            point: 2D location (vertex), z-axis of 3D vertices is ignored.
            start_width: start width of next segment
            end_width:  end width of next segment

        """
        if end_width is None:
            end_width = start_width
        point = Vec2(point)
        stations = self._stations

        if bool(stations) and stations[-1].vertex.isclose(point, abs_tol=self.abs_tol):
            # replace last station
            stations.pop()
        stations.append(LinearStation(point, float(start_width), float(end_width)))

    def faces(self) -> Iterable[Face]:
        """Yields all faces as 4-tuples of :class:`~ezdxf.math.Vec2` objects.

        First and last miter is 90 degrees if the path is not closed, otherwise
        the intersection of first and last segment is taken into account,
        a closed path has to have explicit the same last and first vertex.

        """
        stations = self._stations
        count = len(stations)
        if count < 2:  # Two or more stations required to create faces
            return

        def offset_rays(
            segment: int,
        ) -> tuple[ConstructionRay, ConstructionRay]:
            """Create offset rays from segment offset vertices."""

            def ray(v1, v2):
                if v1.isclose(v2):
                    # vertices too close to define a ray, offset ray is parallel to segment:
                    angle = (
                        stations[segment].vertex - stations[segment + 1].vertex
                    ).angle
                    return ConstructionRay(v1, angle)
                else:
                    return ConstructionRay(v1, v2)

            left1, left2, right1, right2 = segments[segment]
            return ray(left1, left2), ray(right1, right2)

        def intersect(
            ray1: ConstructionRay, ray2: ConstructionRay, default: Vec2
        ) -> Vec2:
            """Intersect two rays but take parallel rays into account."""
            # check for nearly parallel rays pi/100 ~1.8 degrees
            angle = abs(ray1.direction.angle_between(ray2.direction))
            if angle < 0.031415 or abs(math.pi - angle) < 0.031415:
                return default
            try:
                return ray1.intersect(ray2)
            except ParallelRaysError:
                return default

        # Path has to be explicit closed by vertices:
        is_closed = stations[0].vertex.isclose(stations[-1].vertex)

        segments = []
        # Each segment has 4 offset vertices normal to the line from start- to
        # end vertex
        # 1st vertex left of line at the start, distance = start_width/2
        # 2nd vertex left of line at the end, distance = end_width/2
        # 3rd vertex right of line at the start, distance = start_width/2
        # 4th vertex right of line at the end, distance = end_width/2
        for station in range(count - 1):
            start_vertex, start_width, end_width = stations[station]
            end_vertex = stations[station + 1].vertex
            # Start- and end vertex are never to close together, close stations
            # will be merged in method LinearTrace.add_station().
            segments.append(
                _normal_offset_points(start_vertex, end_vertex, start_width, end_width)
            )

        # offset rays:
        # 1 is the upper or left of line
        # 2 is the lower or right of line
        offset_ray1, offset_ray2 = offset_rays(0)
        prev_offset_ray1 = None
        prev_offset_ray2 = None

        # Store last vertices explicit, they get modified for closed paths.
        last_up1, last_up2, last_down1, last_down2 = segments[-1]

        for i in range(len(segments)):
            up1, up2, down1, down2 = segments[i]
            if i == 0:
                # Set first vertices of the first face.
                if is_closed:
                    # Compute first two vertices as intersection of first and
                    # last segment
                    last_offset_ray1, last_offset_ray2 = offset_rays(len(segments) - 1)
                    vtx0 = intersect(last_offset_ray1, offset_ray1, up1)
                    vtx1 = intersect(last_offset_ray2, offset_ray2, down1)

                    # Store last vertices for the closing face.
                    last_up2 = vtx0
                    last_down2 = vtx1
                else:
                    # Set first two vertices of the first face for an open path.
                    vtx0 = up1
                    vtx1 = down1
                prev_offset_ray1 = offset_ray1
                prev_offset_ray2 = offset_ray2
            else:
                # Compute first two vertices for the actual face.
                vtx0 = intersect(prev_offset_ray1, offset_ray1, up1)  # type: ignore
                vtx1 = intersect(prev_offset_ray2, offset_ray2, down1)  # type: ignore

            if i < len(segments) - 1:
                # Compute last two vertices for the actual face.
                next_offset_ray1, next_offset_ray2 = offset_rays(i + 1)
                vtx2 = intersect(next_offset_ray2, offset_ray2, down2)
                vtx3 = intersect(next_offset_ray1, offset_ray1, up2)
                prev_offset_ray1 = offset_ray1
                prev_offset_ray2 = offset_ray2
                offset_ray1 = next_offset_ray1
                offset_ray2 = next_offset_ray2
            else:
                # Pickup last two vertices for the last face.
                vtx2 = last_down2
                vtx3 = last_up2
            yield vtx0, vtx1, vtx2, vtx3


def _normal_offset_points(
    start: Vec2, end: Vec2, start_width: float, end_width: float
) -> Face:
    dir_vector = (end - start).normalize()
    ortho = dir_vector.orthogonal(True)
    offset_start = ortho.normalize(start_width / 2)
    offset_end = ortho.normalize(end_width / 2)
    return (
        start + offset_start,
        end + offset_end,
        start - offset_start,
        end - offset_end,
    )


_NULLVEC2 = Vec2((0, 0))


class CurvedTrace(AbstractTrace):
    """2D banded curves like arcs or splines with start- and end width.

    Represents always only one curved entity and all miter of curve segments
    are perpendicular to curve tangents.

    Accepts 3D input, but z-axis is ignored.

    """

    def __init__(self) -> None:
        self._stations: list[CurveStation] = []

    def __len__(self):
        return len(self._stations)

    def __getitem__(self, item):
        return self._stations[item]

    @classmethod
    def from_spline(
        cls,
        spline: BSpline,
        start_width: float,
        end_width: float,
        segments: int,
    ) -> CurvedTrace:
        """
        Create curved trace from a B-spline.

        Args:
            spline: :class:`~ezdxf.math.BSpline` object
            start_width: start width
            end_width: end width
            segments: count of segments for approximation

        """
        curve_trace = cls()
        count = segments + 1
        t = np.linspace(0, spline.max_t, count)
        for (point, derivative), width in zip(
            spline.derivatives(t, n=1), np.linspace(start_width, end_width, count)
        ):
            normal = Vec2(derivative).orthogonal(True)
            curve_trace._append(Vec2(point), normal, width)  # type: ignore
        return curve_trace

    @classmethod
    def from_arc(
        cls,
        arc: ConstructionArc,
        start_width: float,
        end_width: float,
        segments: int = 64,
    ) -> CurvedTrace:
        """
        Create curved trace from an arc.

        Args:
            arc: :class:`~ezdxf.math.ConstructionArc` object
            start_width: start width
            end_width: end width
            segments: count of segments for full circle (360 degree)
                approximation, partial arcs have proportional less segments,
                but at least 3

        Raises:
            ValueError: if arc.radius <= 0

        """
        if arc.radius <= 0:
            raise ValueError(f"Invalid radius: {arc.radius}.")
        curve_trace = cls()
        count = max(math.ceil(arc.angle_span / 360.0 * segments), 3) + 1
        center = Vec2(arc.center)
        for point, width in zip(
            arc.vertices(arc.angles(count)),
            np.linspace(start_width, end_width, count),
        ):
            curve_trace._append(point, point - center, width)  # type: ignore
        return curve_trace

    def _append(self, point: Vec2, normal: Vec2, width: float) -> None:
        """
        Add a curve trace station (like a vertex) at location `point`.

        Args:
            point: 2D curve location (vertex), z-axis of 3D vertices is ignored.
            normal: curve normal
            width:  width of station

        """
        if _NULLVEC2.isclose(normal):
            normal = _NULLVEC2
        else:
            normal = normal.normalize(width / 2)
        self._stations.append(CurveStation(point + normal, point - normal))

    def faces(self) -> Iterable[Face]:
        """Yields all faces as 4-tuples of :class:`~ezdxf.math.Vec2` objects."""
        count = len(self._stations)
        if count < 2:  # Two or more stations required to create faces
            return

        vtx0 = None
        vtx1 = None
        for vtx2, vtx3 in self._stations:
            if vtx0 is None:
                vtx0 = vtx3
                vtx1 = vtx2
                continue
            yield vtx0, vtx1, vtx2, vtx3
            vtx0 = vtx3
            vtx1 = vtx2


def should_include_vertex_for_rendering(vertex_flags: int) -> bool:
    """Determine if a vertex should be included in rendering based on its flags.

    According to the DXF specification:
    - Flag 0: Straight-segment vertex (include)
    - Flag 1: Extra curve-fit vertex (exclude - auto-generated)
    - Flag 2: Curve-fit tangent vertex (include - carries tangent info)
    - Flag 8: Spline-vertex (fit-point) for 2D spline (include)
    - Flag 16: Spline-frame control point (exclude - internal calculation only)

    Args:
        vertex_flags: The vertex flags value (group code 70)

    Returns:
        True if the vertex should be included in rendering, False otherwise
    """
    # Exclude extra vertices created by curve-fitting (flag 1)
    if vertex_flags & VTX_EXTRA_VERTEX_CREATED:
        return False

    # Exclude spline frame control points (flag 16)
    if vertex_flags & VTX_SPLINE_FRAME_CONTROL_POINT:
        return False

    # Include all other vertex types:
    # - Flag 0: Regular straight-segment vertices
    # - Flag 2: Curve-fit tangent vertices (carry tangent direction)
    # - Flag 8: Spline fit-points
    return True


class TraceBuilder(Sequence):
    """Sequence of 2D banded lines like polylines with start- and end width or
    curves with start- and end width.


    .. note::

        Accepts 3D input, but z-axis is ignored. The :class:`TraceBuilder` is a
        2D only object and uses only the :ref:`OCS` coordinates!

    """

    def __init__(self) -> None:
        self._traces: list[AbstractTrace] = []
        self.abs_tol = 1e-12

    def __len__(self):
        return len(self._traces)

    def __getitem__(self, item):
        return self._traces[item]

    def append(self, trace: AbstractTrace) -> None:
        """Append a new trace."""
        self._traces.append(trace)

    def faces(self) -> Iterable[Face]:
        """Yields all faces as 4-tuples of :class:`~ezdxf.math.Vec2` objects
        in :ref:`OCS`.
        """
        for trace in self._traces:
            yield from trace.faces()

    def faces_wcs(self, ocs: OCS, elevation: float) -> Iterable[Sequence[Vec3]]:
        """Yields all faces as 4-tuples of :class:`~ezdxf.math.Vec3` objects
        in :ref:`WCS`.
        """
        for face in self.faces():
            yield tuple(ocs.points_to_wcs(Vec3(v.x, v.y, elevation) for v in face))

    def polygons(self) -> Iterable[Polygon]:
        """Yields for each sub-trace a single polygon as sequence of
        :class:`~ezdxf.math.Vec2` objects in :ref:`OCS`.
        """
        for trace in self._traces:
            yield trace.polygon()

    def polygons_wcs(self, ocs: OCS, elevation: float) -> Iterable[Sequence[Vec3]]:
        """Yields for each sub-trace a single polygon as sequence of
        :class:`~ezdxf.math.Vec3` objects in :ref:`WCS`.
        """
        for trace in self._traces:
            yield tuple(
                ocs.points_to_wcs(Vec3(v.x, v.y, elevation) for v in trace.polygon())
            )

    def virtual_entities(
        self, dxftype="TRACE", dxfattribs=None, doc: Optional[Drawing] = None
    ) -> Iterable[Quadrilateral]:
        """Yields faces as SOLID, TRACE or 3DFACE entities with DXF attributes
        given in `dxfattribs`.

        If a document is given, the doc attribute of the new entities will be
        set and the new entities will be automatically added to the entity
        database of that document.

        .. note::

            The :class:`TraceBuilder` is a 2D only object and uses only the
            :ref:`OCS` coordinates!

        Args:
            dxftype: DXF type as string, "SOLID", "TRACE" or "3DFACE"
            dxfattribs: DXF attributes for SOLID, TRACE or 3DFACE entities
            doc: associated document

        """
        for trace in self._traces:
            yield from trace.virtual_entities(dxftype, dxfattribs, doc)

    def close(self):
        """Close multi traces by merging first and last trace, if linear traces."""
        traces = self._traces
        if len(traces) < 2:
            return
        if isinstance(traces[0], LinearTrace) and isinstance(traces[-1], LinearTrace):
            first = cast(LinearTrace, traces.pop(0))
            last = cast(LinearTrace, traces[-1])
            for point, start_width, end_width in first:
                last.add_station(point, start_width, end_width)

    @classmethod
    def from_polyline(cls, polyline: DXFGraphic, segments: int = 64) -> TraceBuilder:
        """
        Create a complete trace from a LWPOLYLINE or a 2D POLYLINE entity, the
        trace consist of multiple sub-traces if :term:`bulge` values are
        present. Uses only the :ref:`OCS` coordinates!

        Args:
            polyline: :class:`~ezdxf.entities.LWPolyline` or 2D
                :class:`~ezdxf.entities.Polyline`
            segments: count of segments for bulge approximation, given count is
                for a full circle, partial arcs have proportional less segments,
                but at least 3

        """
        dxftype = polyline.dxftype()
        if dxftype == "LWPOLYLINE":
            polyline = cast("LWPolyline", polyline)
            const_width = polyline.dxf.const_width
            points = []
            for x, y, start_width, end_width, bulge in polyline.lwpoints:
                location = Vec2(x, y)
                if const_width:
                    # This is AutoCAD behavior, BricsCAD uses const width
                    # only for missing width values.
                    start_width = const_width
                    end_width = const_width
                points.append((location, start_width, end_width, bulge))
            closed = polyline.closed
        elif dxftype == "POLYLINE":
            polyline = cast("Polyline", polyline)
            if not polyline.is_2d_polyline:
                raise TypeError("2D POLYLINE required")
            closed = polyline.is_closed
            default_start_width = polyline.dxf.default_start_width
            default_end_width = polyline.dxf.default_end_width

            filtered_vertices = [
                vertex for vertex in polyline.vertices
                if should_include_vertex_for_rendering(vertex.dxf.get("flags", 0))
            ]

            points = []
            for vertex in filtered_vertices:
                location = Vec2(vertex.dxf.location)
                if vertex.dxf.hasattr("start_width"):
                    start_width = vertex.dxf.start_width
                else:
                    start_width = default_start_width
                if vertex.dxf.hasattr("end_width"):
                    end_width = vertex.dxf.end_width
                else:
                    end_width = default_end_width
                bulge = vertex.dxf.bulge
                points.append((location, start_width, end_width, bulge))
        else:
            raise TypeError(f"Invalid DXF type {dxftype}")

        if closed and not points[0][0].isclose(points[-1][0]):
            # close polyline explicit
            points.append(points[0])

        trace = cls()
        store_bulge = 0.0
        store_start_width = 0.0
        store_end_width = 0.0
        store_point: UVec | None = None

        linear_trace = LinearTrace()
        for point, start_width, end_width, bulge in points:
            if store_bulge != 0.0:
                center, start_angle, end_angle, radius = bulge_to_arc(
                    store_point, point, store_bulge
                )
                if radius > 0:
                    arc = ConstructionArc(
                        center,
                        radius,
                        math.degrees(start_angle),
                        math.degrees(end_angle),
                        is_counter_clockwise=True,
                    )
                    if arc.start_point.isclose(point):
                        sw = store_end_width
                        ew = store_start_width
                    else:
                        ew = store_end_width
                        sw = store_start_width
                    trace.append(CurvedTrace.from_arc(arc, sw, ew, segments))
                store_bulge = 0.0

            if bulge != 0.0:  # arc from prev_point to point
                if linear_trace.is_started:
                    linear_trace.add_station(point, start_width, end_width)
                    trace.append(linear_trace)
                    linear_trace = LinearTrace()
                store_bulge = bulge
                store_start_width = start_width
                store_end_width = end_width
                store_point = point
                continue

            linear_trace.add_station(point, start_width, end_width)
        if linear_trace.is_started:
            trace.append(linear_trace)

        if closed and len(trace) > 1:
            # This is required for traces with multiple paths to create the correct
            # miter at the closing point. (only linear to linear trace).
            trace.close()
        return trace



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import NamedTuple, TYPE_CHECKING
from typing_extensions import Self

import math
import enum
import dataclasses
from ezdxf.math import Vec2, BoundingBox2d, Matrix44

if TYPE_CHECKING:
    from ezdxf.layouts.layout import Layout as DXFLayout


class Units(enum.IntEnum):
    """Page units as enum.

    Attributes:
        inch: 25.4 mm
        px: 1/96 inch
        pt: 1/72 inch
        mm:
        cm:

    """

    # equivalent to ezdxf.units if possible
    inch = 1
    px = 2  # no equivalent DXF unit
    pt = 3  # no equivalent DXF unit
    mm = 4
    cm = 5


# all page sizes in landscape orientation
PAGE_SIZES = {
    "ISO A0": (1189, 841, Units.mm),
    "ISO A1": (841, 594, Units.mm),
    "ISO A2": (594, 420, Units.mm),
    "ISO A3": (420, 297, Units.mm),
    "ISO A4": (297, 210, Units.mm),
    "ANSI A": (11, 8.5, Units.inch),
    "ANSI B": (17, 11, Units.inch),
    "ANSI C": (22, 17, Units.inch),
    "ANSI D": (34, 22, Units.inch),
    "ANSI E": (44, 34, Units.inch),
    "ARCH C": (24, 18, Units.inch),
    "ARCH D": (36, 24, Units.inch),
    "ARCH E": (48, 36, Units.inch),
    "ARCH E1": (42, 30, Units.inch),
    "Letter": (11, 8.5, Units.inch),
    "Legal": (14, 8.5, Units.inch),
}


UNITS_TO_MM = {
    Units.mm: 1.0,
    Units.cm: 10.0,
    Units.inch: 25.4,
    Units.px: 25.4 / 96.0,
    Units.pt: 25.4 / 72.0,
}


class PageAlignment(enum.IntEnum):
    """Page alignment of content as enum.

    Attributes:
        TOP_LEFT:
        TOP_CENTER:
        TOP_RIGHT:
        MIDDLE_LEFT:
        MIDDLE_CENTER:
        MIDDLE_RIGHT:
        BOTTOM_LEFT:
        BOTTOM_CENTER:
        BOTTOM_RIGHT:

    """

    TOP_LEFT = 1
    TOP_CENTER = 2
    TOP_RIGHT = 3
    MIDDLE_LEFT = 4
    MIDDLE_CENTER = 5
    MIDDLE_RIGHT = 6
    BOTTOM_LEFT = 7
    BOTTOM_CENTER = 8
    BOTTOM_RIGHT = 9


class Margins(NamedTuple):
    """Page margins definition class

    Attributes:
        top:
        left:
        bottom:
        right:

    """

    top: float
    right: float
    bottom: float
    left: float

    @classmethod
    def all(cls, margin: float) -> Self:
        """Returns a page margins definition class with four equal margins."""
        return cls(margin, margin, margin, margin)

    @classmethod
    def all2(cls, top_bottom: float, left_right: float) -> Self:
        """Returns a page margins definition class with equal top-bottom and
        left-right margins.
        """
        return cls(top_bottom, left_right, top_bottom, left_right)

    # noinspection PyArgumentList
    def scale(self, factor: float) -> Self:
        return self.__class__(
            self.top * factor,
            self.right * factor,
            self.bottom * factor,
            self.left * factor,
        )


@dataclasses.dataclass
class Page:
    """Page definition class

    Attributes:

        width: page width, 0 for auto-detect
        height: page height, 0 for auto-detect
        units: page units as enum :class:`Units`
        margins: page margins in page units
        max_width: limit width for auto-detection, 0 for unlimited
        max_height: limit height for auto-detection, 0 for unlimited

    """

    width: float
    height: float
    units: Units = Units.mm
    margins: Margins = Margins.all(0)
    max_width: float = 0.0
    max_height: float = 0.0

    def __post_init__(self):
        assert isinstance(self.units, Units), "units require type <Units>"
        assert isinstance(self.margins, Margins), "margins require type <Margins>"

    @property
    def to_mm_factor(self) -> float:
        return UNITS_TO_MM[self.units]

    @property
    def width_in_mm(self) -> float:
        """Returns the page width in mm."""
        return round(self.width * self.to_mm_factor, 1)

    @property
    def max_width_in_mm(self) -> float:
        """Returns max page width in mm."""
        return round(self.max_width * self.to_mm_factor, 1)

    @property
    def height_in_mm(self) -> float:
        """Returns the page height in mm."""
        return round(self.height * self.to_mm_factor, 1)

    @property
    def max_height_in_mm(self) -> float:
        """Returns max page height in mm."""
        return round(self.max_height * self.to_mm_factor, 1)

    @property
    def margins_in_mm(self) -> Margins:
        """Returns the page margins in mm."""
        return self.margins.scale(self.to_mm_factor)

    @property
    def is_landscape(self) -> bool:
        """Returns ``True`` if the page has landscape orientation."""
        return self.width > self.height

    @property
    def is_portrait(self) -> bool:
        """Returns ``True`` if the page has portrait orientation. (square is portrait)"""
        return self.width <= self.height

    def to_landscape(self) -> None:
        """Converts the page to landscape orientation."""
        if self.is_portrait:
            self.width, self.height = self.height, self.width

    def to_portrait(self) -> None:
        """Converts the page to portrait orientation."""
        if self.is_landscape:
            self.width, self.height = self.height, self.width

    def get_margin_rect(self, top_origin=True) -> tuple[Vec2, Vec2]:
        """Returns the bottom-left and the top-right corner of the page margins in mm.
        The origin (0, 0) is the top-left corner of the page if `top_origin` is
        ``True`` or in the bottom-left corner otherwise.
        """
        margins = self.margins_in_mm
        right_margin = self.width_in_mm - margins.right
        page_height = self.height_in_mm
        if top_origin:
            bottom_left = Vec2(margins.left, margins.top)
            top_right = Vec2(right_margin, page_height - margins.bottom)
        else:  # bottom origin
            bottom_left = Vec2(margins.left, margins.bottom)
            top_right = Vec2(right_margin, page_height - margins.top)
        return bottom_left, top_right

    @classmethod
    def from_dxf_layout(cls, layout: DXFLayout) -> Self:
        """Returns the :class:`Page` based on the DXF attributes stored in the LAYOUT 
        entity. The modelspace layout often **doesn't** have usable page settings!

        Args:
            layout: any paperspace layout or the modelspace layout

        """
        # all layout measurements in mm
        width = round(layout.dxf.paper_width, 1)
        height = round(layout.dxf.paper_height, 1)
        top = round(layout.dxf.top_margin, 1)
        right = round(layout.dxf.right_margin, 1)
        bottom = round(layout.dxf.bottom_margin, 1)
        left = round(layout.dxf.left_margin, 1)

        rotation = layout.dxf.plot_rotation
        if rotation == 1:  # 90 degrees
            return cls(
                height,
                width,
                Units.mm,
                margins=Margins(top=right, right=bottom, bottom=left, left=top),
            )
        elif rotation == 2:  # 180 degrees
            return cls(
                width,
                height,
                Units.mm,
                margins=Margins(top=bottom, right=left, bottom=top, left=right),
            )
        elif rotation == 3:  # 270 degrees
            return cls(
                height,
                width,
                Units.mm,
                margins=Margins(top=left, right=top, bottom=right, left=bottom),
            )
        return cls(  # 0 degrees
            width,
            height,
            Units.mm,
            margins=Margins(top=top, right=right, bottom=bottom, left=left),
        )


@dataclasses.dataclass
class Settings:
    """The Layout settings.

    Attributes:
        content_rotation: Rotate content about 0, 90,  180 or 270 degrees
        fit_page: Scale content to fit the page.
        page_alignment: Supported by backends that use the :class:`Page` class to define
            the size of the output media, default alignment is :attr:`PageAlignment.MIDDLE_CENTER`
        crop_at_margins: crops the content at the page margins if ``True``, when
            supported by the backend, default is ``False``
        scale: Factor to scale the DXF units of model- or paperspace, to represent 1mm
            in the rendered output drawing. Only uniform scaling is supported.

            e.g. scale 1:100 and DXF units are meters, 1m = 1000mm corresponds 10mm in
            the output drawing = 10 / 1000 = 0.01;

            e.g. scale 1:1; DXF units are mm = 1 / 1 = 1.0 the default value

            The value is ignored if the page size is defined and the content fits the page and
            the value is also used to determine missing page sizes (width or height).
        max_stroke_width: Used for :class:`LineweightPolicy.RELATIVE` policy,
            :attr:`max_stroke_width` is defined as percentage of the content extents,
            e.g. 0.001 is 0.1% of max(page-width, page-height)
        min_stroke_width: Used for :class:`LineweightPolicy.RELATIVE` policy,
            :attr:`min_stroke_width` is defined as percentage of :attr:`max_stroke_width`,
            e.g. 0.05 is 5% of :attr:`max_stroke_width`
        fixed_stroke_width: Used for :class:`LineweightPolicy.RELATIVE_FIXED` policy,
            :attr:`fixed_stroke_width` is defined as percentage of :attr:`max_stroke_width`,
            e.g. 0.15 is 15% of :attr:`max_stroke_width`
        output_coordinate_space: expert feature to map the DXF coordinates to the
            output coordinate system [0, output_coordinate_space]
        output_layers: For supported backends, separate the entities into 'layers' in the output

    """

    content_rotation: int = 0
    fit_page: bool = True
    scale: float = 1.0
    page_alignment: PageAlignment = PageAlignment.MIDDLE_CENTER
    crop_at_margins: bool = False
    # for LineweightPolicy.RELATIVE
    # max_stroke_width is defined as percentage of the content extents
    max_stroke_width: float = 0.001  # 0.1% of max(width, height) in viewBox coords
    # min_stroke_width is defined as percentage of max_stroke_width
    min_stroke_width: float = 0.05  # 5% of max_stroke_width
    # StrokeWidthPolicy.fixed_1
    # fixed_stroke_width is defined as percentage of max_stroke_width
    fixed_stroke_width: float = 0.15  # 15% of max_stroke_width
    # PDF, HPGL expect the coordinates in the first quadrant and SVG has an inverted
    # y-axis, so transformation from DXF to the output coordinate system is required.
    # The output_coordinate_space defines the space into which the DXF coordinates are
    # mapped, the range is [0, output_coordinate_space] for the larger page
    # dimension - aspect ratio is always preserved - these are CAD drawings!
    # The SVGBackend uses this feature to map all coordinates to integer values:
    output_coordinate_space: float = 1_000_000  # recommended value for the SVGBackend
    output_layers: bool = True

    def __post_init__(self) -> None:
        if self.content_rotation not in (0, 90, 180, 270):
            raise ValueError(
                f"invalid content rotation {self.content_rotation}, "
                f"expected: 0, 90, 180, 270"
            )

    def page_output_scale_factor(self, page: Page) -> float:
        """Returns the scaling factor to map page coordinates in mm to output space
        coordinates.
        """
        try:
            return self.output_coordinate_space / max(
                page.width_in_mm, page.height_in_mm
            )
        except ZeroDivisionError:
            return 1.0


class Layout:
    def __init__(self, render_box: BoundingBox2d, flip_y=False) -> None:
        super().__init__()
        self.flip_y: float = -1.0 if flip_y else 1.0
        self.render_box = render_box

    def get_rotation(self, settings: Settings) -> int:
        if settings.content_rotation not in (0, 90, 180, 270):
            raise ValueError("content rotation must be 0, 90, 180 or 270 degrees")
        rotation = settings.content_rotation
        if self.flip_y == -1.0:
            if rotation == 90:
                rotation = 270
            elif rotation == 270:
                rotation = 90
        return rotation

    def get_content_size(self, rotation: int) -> Vec2:
        content_size = self.render_box.size
        if rotation in (90, 270):
            # swap x, y to apply rotation to content_size
            content_size = Vec2(content_size.y, content_size.x)
        return content_size

    def get_final_page(self, page: Page, settings: Settings) -> Page:
        rotation = self.get_rotation(settings)
        content_size = self.get_content_size(rotation)
        return final_page_size(content_size, page, settings)

    def get_placement_matrix(
        self, page: Page, settings=Settings(), top_origin=True
    ) -> Matrix44:
        # Argument `page` has to be the resolved final page size!
        rotation = self.get_rotation(settings)

        content_size = self.get_content_size(rotation)
        content_size_mm = content_size * settings.scale
        if settings.fit_page:
            content_size_mm *= fit_to_page(content_size_mm, page)
        try:
            scale_dxf_to_mm = content_size_mm.x / content_size.x
        except ZeroDivisionError:
            scale_dxf_to_mm = 1.0
        # map output coordinates to range [0, output_coordinate_space]
        scale_mm_to_output_space = settings.page_output_scale_factor(page)
        scale = scale_dxf_to_mm * scale_mm_to_output_space
        m = placement_matrix(
            self.render_box,
            sx=scale,
            sy=scale * self.flip_y,
            rotation=rotation,
            page=page,
            output_coordinate_space=settings.output_coordinate_space,
            page_alignment=settings.page_alignment,
            top_origin=top_origin,
        )
        return m


def final_page_size(content_size: Vec2, page: Page, settings: Settings) -> Page:
    scale = settings.scale
    width = page.width_in_mm
    height = page.height_in_mm
    margins = page.margins_in_mm
    if width == 0.0:
        width = scale * content_size.x + margins.left + margins.right
    if height == 0.0:
        height = scale * content_size.y + margins.top + margins.bottom

    width, height = limit_page_size(
        width, height, page.max_width_in_mm, page.max_height_in_mm
    )
    return Page(round(width, 1), round(height, 1), Units.mm, margins)


def limit_page_size(
    width: float, height: float, max_width: float, max_height: float
) -> tuple[float, float]:
    try:
        ar = width / height
    except ZeroDivisionError:
        return width, height
    if max_height:
        height = min(max_height, height)
        width = height * ar
    if max_width and width > max_width:
        width = min(max_width, width)
        height = width / ar
    return width, height


def fit_to_page(content_size_mm: Vec2, page: Page) -> float:
    margins = page.margins_in_mm
    try:
        sx = (page.width_in_mm - margins.left - margins.right) / content_size_mm.x
        sy = (page.height_in_mm - margins.top - margins.bottom) / content_size_mm.y
    except ZeroDivisionError:
        return 1.0
    return min(sx, sy)


def placement_matrix(
    bbox: BoundingBox2d,
    sx: float,
    sy: float,
    rotation: float,
    page: Page,
    output_coordinate_space: float,
    page_alignment: PageAlignment = PageAlignment.MIDDLE_CENTER,
    # top_origin True: page origin (0, 0) in top-left corner, +y axis pointing down
    # top_origin False: page origin (0, 0) in bottom-left corner, +y axis pointing up
    top_origin=True,
) -> Matrix44:
    """Returns a matrix to place the bbox in the first quadrant of the coordinate
    system (+x, +y).
    """
    try:
        scale_mm_to_vb = output_coordinate_space / max(
            page.width_in_mm, page.height_in_mm
        )
    except ZeroDivisionError:
        scale_mm_to_vb = 1.0
    margins = page.margins_in_mm

    # create scaling and rotation matrix:
    if abs(sx) < 1e-9:
        sx = 1.0
    if abs(sy) < 1e-9:
        sy = 1.0
    m = Matrix44.scale(sx, sy, 1.0)
    if rotation:
        m @= Matrix44.z_rotate(math.radians(rotation))

    # calc bounding box of the final output canvas:
    corners = m.transform_vertices(bbox.rect_vertices())
    canvas = BoundingBox2d(corners)

    # shift content to first quadrant +x/+y
    tx, ty = canvas.extmin

    # align content within margins
    view_box_content_x = (
        page.width_in_mm - margins.left - margins.right
    ) * scale_mm_to_vb
    view_box_content_y = (
        page.height_in_mm - margins.top - margins.bottom
    ) * scale_mm_to_vb
    dx = view_box_content_x - canvas.size.x
    dy = view_box_content_y - canvas.size.y
    offset_x = margins.left * scale_mm_to_vb  # left
    if top_origin:
        offset_y = margins.top * scale_mm_to_vb
    else:
        offset_y = margins.bottom * scale_mm_to_vb

    if is_center_aligned(page_alignment):
        offset_x += dx / 2
    elif is_right_aligned(page_alignment):
        offset_x += dx
    if is_middle_aligned(page_alignment):
        offset_y += dy / 2
    elif is_bottom_aligned(page_alignment):
        if top_origin:
            offset_y += dy
    else:  # top aligned
        if not top_origin:
            offset_y += dy
    return m @ Matrix44.translate(-tx + offset_x, -ty + offset_y, 0)


def is_left_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.TOP_LEFT,
        PageAlignment.MIDDLE_LEFT,
        PageAlignment.BOTTOM_LEFT,
    )


def is_center_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.TOP_CENTER,
        PageAlignment.MIDDLE_CENTER,
        PageAlignment.BOTTOM_CENTER,
    )


def is_right_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.TOP_RIGHT,
        PageAlignment.MIDDLE_RIGHT,
        PageAlignment.BOTTOM_RIGHT,
    )


def is_top_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.TOP_LEFT,
        PageAlignment.TOP_CENTER,
        PageAlignment.TOP_RIGHT,
    )


def is_middle_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.MIDDLE_LEFT,
        PageAlignment.MIDDLE_CENTER,
        PageAlignment.MIDDLE_RIGHT,
    )


def is_bottom_aligned(align: PageAlignment) -> bool:
    return align in (
        PageAlignment.BOTTOM_LEFT,
        PageAlignment.BOTTOM_CENTER,
        PageAlignment.BOTTOM_RIGHT,
    )



# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf.entities.dxfobj import SortEntsTable


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


def get_entry(table, index):
    return list(table)[index]


def test_sort_entities_table(doc):
    sort_entities_table = doc.objects.new_entity(
        "SORTENTSTABLE", {"block_record_handle": "FFFF"}
    )
    assert sort_entities_table.dxftype() == "SORTENTSTABLE"
    assert sort_entities_table.dxf.block_record_handle == "FFFF"
    assert len(sort_entities_table) == 0
    sort_entities_table.append("AAA", "BBB")
    assert get_entry(sort_entities_table, 0) == ("AAA", "BBB")


def test_sort_entities_table_as_list(doc):
    sort_entities_table = doc.objects.new_entity("SORTENTSTABLE", {})
    sort_entities_table.set_handles(
        [
            ("AAA", "BBB"),
            ("CCC", "DDD"),
            ("EEE", "FFF"),
        ]
    )
    assert len(sort_entities_table) == 3
    assert get_entry(sort_entities_table, 0) == ("AAA", "BBB")
    assert get_entry(sort_entities_table, -1) == ("EEE", "FFF")

    sort_entities_table.clear()
    assert len(sort_entities_table) == 0


def test_sort_entities_table_to_dict(doc):
    sort_entities_table = doc.objects.new_entity("SORTENTSTABLE", {})
    sort_entities_table.set_handles(
        [
            ("AAA", "BBB"),
            ("CCC", "DDD"),
            ("EEE", "FFF"),
        ]
    )
    assert len(sort_entities_table) == 3
    assert get_entry(sort_entities_table, 2) == ("EEE", "FFF")

    # simple way to dict()
    d = dict(sort_entities_table)
    assert d["AAA"] == "BBB"
    assert d["CCC"] == "DDD"
    assert d["EEE"] == "FFF"


def test_remove_invalid_handles(doc):
    sort_entities_table = doc.objects.new_entity("SORTENTSTABLE", {})
    sort_entities_table.set_handles(
        [
            ("AAA", "BBB"),
            ("CCC", "DDD"),
            ("EEE", "FFF"),
        ]
    )
    assert len(sort_entities_table) == 3
    sort_entities_table.remove_invalid_handles()
    assert len(sort_entities_table) == 0


def test_remove_handle(doc):
    sort_entities_table = doc.objects.new_entity("SORTENTSTABLE", {})
    sort_entities_table.set_handles(
        [
            ("AAA", "BBB"),
            ("CCC", "DDD"),
            ("EEE", "FFF"),
        ]
    )
    assert len(sort_entities_table) == 3
    sort_entities_table.remove_handle("AAA")
    assert len(sort_entities_table) == 2
    # no exception if handle not exists
    sort_entities_table.remove_handle("FFFF")
    assert len(sort_entities_table) == 2


SORT_ENTITIES_TABLE = """0
SORTENTSTABLE
5
0
102
{ACAD_REACTORS
330
0
102
}
330
0
100
AcDbSortentsTable
330
ABBA
331
1
5
A
331
2
5
B
"""


def test_load_table():
    table = SortEntsTable.from_text(SORT_ENTITIES_TABLE)
    assert table.dxf.block_record_handle == "ABBA"
    assert len(table) == 2
    assert table.table["1"] == "A"
    assert table.table["2"] == "B"



# Copyright (c) 2019-2020, Manfred Moitzi
# License: MIT License
from ezdxf.math import ConstructionLine, Vec2


class TestConstructionLine:
    def test_is_vertical(self):
        assert ConstructionLine((0, 0), (10, 0)).is_vertical is False
        assert ConstructionLine((5, -5), (5, 5)).is_vertical is True

    def test_left_of_line(self):
        line = ConstructionLine((0, 0), (0.1, 1))
        assert line.is_point_left_of_line(Vec2(-1, 0)) is True
        assert line.is_point_left_of_line(Vec2(1, 0)) is False
        assert line.is_point_left_of_line(Vec2(-1, -1)) is True

        line = ConstructionLine((0, 0), (0, -1))
        assert line.is_point_left_of_line(Vec2(1, 0)) is True

        line = ConstructionLine((0, 0), (-1, 0.1))
        assert line.is_point_left_of_line(Vec2(-1, 0)) is True

        line = ConstructionLine((0, 0), (10, 0))
        assert line.is_point_left_of_line(Vec2(0, 0)) is False
        assert line.is_point_left_of_line(Vec2(10, 0)) is False
        assert line.is_point_left_of_line(Vec2(10, 1)) is True
        assert line.is_point_left_of_line(Vec2(10, -1)) is False

        line = ConstructionLine((10, 0), (0, 0))
        assert line.is_point_left_of_line(Vec2(0, 0)) is False
        assert line.is_point_left_of_line(Vec2(10, 0)) is False
        assert line.is_point_left_of_line(Vec2(10, 1)) is False
        assert line.is_point_left_of_line(Vec2(10, -1)) is True

        line = ConstructionLine((0, 0), (0, 10))
        assert line.is_point_left_of_line(Vec2(0, 0)) is False
        assert line.is_point_left_of_line(Vec2(0, 10)) is False
        assert line.is_point_left_of_line(Vec2(1, 10)) is False
        assert line.is_point_left_of_line(Vec2(-1, 10)) is True

        line = ConstructionLine((0, 10), (0, 0))
        assert line.is_point_left_of_line(Vec2(0, 0)) is False
        assert line.is_point_left_of_line(Vec2(0, 10)) is False
        assert line.is_point_left_of_line(Vec2(1, 10)) is True
        assert line.is_point_left_of_line(Vec2(-1, 10)) is False

    def test_intersect_horizontal_line(self):
        line = ConstructionLine((0, 0), (10, 0))
        assert line.intersect(ConstructionLine((0, 0), (10, 0))) is None
        assert line.intersect(ConstructionLine((0, 1), (10, 1))) is None
        assert line.intersect(ConstructionLine((0, -1), (10, 1))) == (5, 0)
        assert line.intersect(ConstructionLine((5, 5), (5, -5))) == (5, 0)
        assert line.intersect(ConstructionLine((5, 5), (5, 1))) is None
        assert line.intersect(ConstructionLine((0, 0), (5, 5))) == (0, 0)

    def test_intersect_vertical_line(self):
        line = ConstructionLine((0, 0), (0, 10))
        assert line.intersect(ConstructionLine((0, 0), (0, 10))) is None
        assert line.intersect(ConstructionLine((1, 0), (1, 10))) is None
        assert line.intersect(ConstructionLine((-1, 0), (1, 10))) == (0, 5)
        assert line.intersect(ConstructionLine((-1, 0), (1, 0))) == (0, 0)
        assert line.intersect(ConstructionLine((-1, 10), (1, 10))) == (0, 10)
        assert line.intersect(ConstructionLine((-1, 11), (1, 11))) is None

    def test_bounding_box(self):
        line = ConstructionLine((0, 0), (7, 10))
        bbox = line.bounding_box
        assert bbox.extmin == (0, 0)
        assert bbox.extmax == (7, 10)

    def test_translate(self):
        line = ConstructionLine((0, 0), (0, 10))
        line.translate(3, 7)
        assert line.start == (3, 7)
        assert line.end == (3, 17)
        bbox = line.bounding_box
        assert bbox.extmin == (3, 7)
        assert bbox.extmax == (3, 17)



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
"""
Module to export any DXF document as DXF version R12 without modifying the source
document.

.. versionadded:: 1.1

To get the best result use the ODA File Converter add-on::

    from ezdxf.addons import odafc

    odafc.convert("any.dxf", "r12.dxf", version="R12")

"""

from __future__ import annotations
from typing import TYPE_CHECKING, TextIO, Callable, Optional
import os
from io import StringIO
import logging

import ezdxf
from ezdxf import const, proxygraphic, path
from ezdxf.document import Drawing
from ezdxf.entities import (
    BlockRecord,
    DXFEntity,
    DXFTagStorage,
    Ellipse,
    Hatch,
    Insert,
    LWPolyline,
    MPolygon,
    MText,
    Mesh,
    Polyface,
    Polyline,
    Spline,
    Textstyle,
)
from ezdxf.entities.polygon import DXFPolygon
from ezdxf.addons import MTextExplode
from ezdxf.entitydb import EntitySpace
from ezdxf.layouts import BlockLayout, VirtualLayout
from ezdxf.lldxf.tagwriter import TagWriter, AbstractTagWriter
from ezdxf.lldxf.types import DXFTag, TAG_STRING_FORMAT
from ezdxf.math import Z_AXIS, Vec3, NULLVEC
from ezdxf.r12strict import R12NameTranslator
from ezdxf.render import MeshBuilder
from ezdxf.sections.table import TextstyleTable

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

__all__ = ["R12Exporter", "convert", "saveas", "write"]

MAX_SAGITTA = 0.01
logger = logging.getLogger("ezdxf")


def convert(doc: Drawing, *, max_sagitta: float = MAX_SAGITTA) -> Drawing:
    """Export and reload DXF document as DXF version R12.

    Writes the DXF document into a temporary file at the file-system and reloads this
    file by the :func:`ezdxf.readfile` function.
    """
    stream = StringIO()
    exporter = R12Exporter(doc, max_sagitta=max_sagitta)
    exporter.write(stream)
    stream.seek(0)
    return ezdxf.read(stream)


def write(doc: Drawing, stream: TextIO, *, max_sagitta: float = MAX_SAGITTA) -> None:
    """Write a DXF document as DXF version R12 to a text stream. The `max_sagitta`
    argument determines the accuracy of the curve flatting for SPLINE and ELLIPSE
    entities.

    Args:
        doc: DXF document to export
        stream: output stream, use :attr:`doc.encoding` as encoding
        max_sagitta: maximum distance from the center of the curve to the
            center of the line segment between two approximation points to
            determine if a segment should be subdivided.

    """
    exporter = R12Exporter(doc, max_sagitta=max_sagitta)
    exporter.write(stream)


def saveas(
    doc: Drawing, filepath: str | os.PathLike, *, max_sagitta: float = MAX_SAGITTA
) -> None:
    """Write a DXF document as DXF version R12 to a file. The `max_sagitta`
    argument determines the accuracy of the curve flatting for SPLINE and ELLIPSE
    entities.

    Args:
        doc: DXF document to export
        filepath: output filename
        max_sagitta: maximum distance from the center of the curve to the
            center of the line segment between two approximation points to
            determine if a segment should be subdivided.

    """
    with open(filepath, "wt", encoding=doc.encoding, errors="dxfreplace") as stream:
        write(
            doc,
            stream,
            max_sagitta=max_sagitta,
        )


def spline_to_polyline(
    spline: Spline, max_sagitta: float, min_segments: int
) -> Polyline:
    polyline = Polyline.new(
        dxfattribs={
            "layer": spline.dxf.layer,
            "linetype": spline.dxf.linetype,
            "color": spline.dxf.color,
            "flags": const.POLYLINE_3D_POLYLINE,
        }
    )

    polyline.append_vertices(points=spline.flattening(max_sagitta, min_segments))
    polyline.new_seqend()
    return polyline


def ellipse_to_polyline(
    ellipse: Ellipse, max_sagitta: float, min_segments: int
) -> Polyline:
    polyline = Polyline.new(
        dxfattribs={
            "layer": ellipse.dxf.layer,
            "linetype": ellipse.dxf.linetype,
            "color": ellipse.dxf.color,
            "flags": const.POLYLINE_3D_POLYLINE,
        }
    )
    polyline.append_vertices(points=ellipse.flattening(max_sagitta, min_segments))
    polyline.new_seqend()
    return polyline


def lwpolyline_to_polyline(lwpolyline: LWPolyline) -> Polyline:
    polyline = Polyline.new(
        dxfattribs={
            "layer": lwpolyline.dxf.layer,
            "linetype": lwpolyline.dxf.linetype,
            "color": lwpolyline.dxf.color,
        }
    )
    polyline.new_seqend()
    polyline.append_formatted_vertices(lwpolyline.get_points(), format="xyseb")
    if lwpolyline.is_closed:
        polyline.close()
    if lwpolyline.dxf.hasattr("const_width"):
        width = lwpolyline.dxf.const_width
        polyline.dxf.default_start_width = width
        polyline.dxf.default_end_width = width
    extrusion = Vec3(lwpolyline.dxf.extrusion)
    if not extrusion.isclose(Z_AXIS):
        polyline.dxf.extrusion = extrusion
        elevation = lwpolyline.dxf.elevation
        polyline.dxf.elevation = Vec3(0, 0, elevation)
        # Set z-axis of VERTEX.location to elevation?

    return polyline


def mesh_to_polyface_mesh(mesh: Mesh) -> Polyface:
    builder = MeshBuilder.from_mesh(mesh)
    return builder.render_polyface(
        VirtualLayout(),
        dxfattribs={
            "layer": mesh.dxf.layer,
            "linetype": mesh.dxf.linetype,
            "color": mesh.dxf.color,
        },
    )


def get_xpl_block_name(entity: DXFEntity) -> str:
    assert entity.dxf.handle is not None
    return f"EZDXF_XPL_{entity.dxftype()}_{entity.dxf.handle}"


def export_lwpolyline(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity, LWPolyline)
    polyline = lwpolyline_to_polyline(entity)
    if len(polyline.vertices):
        polyline.export_dxf(exporter.tagwriter())


def export_mesh(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity, Mesh)
    polyface_mesh = mesh_to_polyface_mesh(entity)
    if len(polyface_mesh.vertices):
        polyface_mesh.export_dxf(exporter.tagwriter())


def export_spline(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity, Spline)
    polyline = spline_to_polyline(
        entity, exporter.max_sagitta, exporter.min_spline_segments
    )
    if len(polyline.vertices):
        polyline.export_dxf(exporter.tagwriter())


def export_ellipse(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity, Ellipse)
    polyline = ellipse_to_polyline(
        entity, exporter.max_sagitta, exporter.min_ellipse_segments
    )
    if len(polyline.vertices):
        polyline.export_dxf(exporter.tagwriter())


def make_insert(name: str, entity: DXFEntity, location=NULLVEC) -> Insert:
    return Insert.new(
        dxfattribs={
            "name": name,
            "layer": entity.dxf.layer,
            "linetype": entity.dxf.linetype,
            "color": entity.dxf.color,
            "insert": location,
        }
    )


def export_proxy_graphic(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity.proxy_graphic, bytes)
    pg = proxygraphic.ProxyGraphic(entity.proxy_graphic)
    try:
        entities = EntitySpace(pg.virtual_entities())
    except proxygraphic.ProxyGraphicError:
        return
    exporter.export_entity_space(entities)


def export_mtext(exporter: R12Exporter, entity: DXFEntity):
    assert isinstance(entity, MText)
    layout = VirtualLayout()
    exporter.explode_mtext(entity, layout)
    exporter.export_entity_space(layout.entity_space)


def export_virtual_entities(exporter: R12Exporter, entity: DXFEntity):
    layout = VirtualLayout()
    try:
        for e in entity.virtual_entities():  # type: ignore
            layout.add_entity(e)
    except Exception:
        return
    exporter.export_entity_space(layout.entity_space)


def export_pattern_fill(entity: DXFEntity, block: BlockLayout) -> None:
    assert isinstance(entity, DXFPolygon)
    dxfattribs = {
        "layer": entity.dxf.layer,
        "color": entity.dxf.color,
    }
    for start, end in entity.render_pattern_lines():
        block.add_line(start, end, dxfattribs=dxfattribs)


def export_solid_fill(
    entity: DXFPolygon,
    block: BlockLayout,
    max_sagitta: float,
    min_segments: int,
) -> None:
    dxfattribs = {
        "layer": entity.dxf.layer,
        "color": entity.dxf.color,
    }

    extrusion = Vec3(entity.dxf.extrusion)
    if not extrusion.is_null and not extrusion.isclose(Z_AXIS):
        dxfattribs["extrusion"] = extrusion

    # triangulation in OCS coordinates, including elevation and offset values:
    for vertices in entity.triangulate(max_sagitta, min_segments):
        block.add_solid(vertices, dxfattribs=dxfattribs)


def export_hatch(exporter: R12Exporter, entity: DXFEntity) -> None:
    assert isinstance(entity, Hatch)
    # export hatch into an anonymous block
    block = exporter.new_block(entity)
    insert = make_insert(block.name, entity)
    insert.export_dxf(exporter.tagwriter())

    if entity.has_pattern_fill:
        export_pattern_fill(entity, block)
    else:
        export_solid_fill(
            entity, block, exporter.max_sagitta, exporter.min_spline_segments
        )


def export_mpolygon(exporter: R12Exporter, entity: DXFEntity) -> None:
    assert isinstance(entity, MPolygon)
    # export mpolygon into an anonymous block
    block = exporter.new_block(entity)
    insert = make_insert(block.name, entity)
    insert.export_dxf(exporter.tagwriter())

    # elevation is the z-axis of the vertices
    path.render_polylines2d(
        block,
        path.from_hatch_ocs(entity, offset=Vec3(entity.dxf.offset)),
        distance=exporter.max_sagitta,
        segments=exporter.min_spline_segments,
        extrusion=Vec3(entity.dxf.extrusion),
        dxfattribs={
            "layer": entity.dxf.layer,
            "linetype": entity.dxf.linetype,
            "color": entity.dxf.color,
        },
    )
    if entity.has_pattern_fill:
        export_pattern_fill(entity, block)
    else:
        export_solid_fill(
            entity, block, exporter.max_sagitta, exporter.min_spline_segments
        )


def export_acad_table(exporter: R12Exporter, entity: DXFEntity) -> None:
    from ezdxf.entities.acad_table import AcadTableBlockContent

    assert isinstance(entity, AcadTableBlockContent)
    table: AcadTableBlockContent = entity
    location = table.get_insert_location()
    block_name = table.get_block_name()
    if not block_name.startswith("*T"):
        return
    try:
        acdb_entity = table.xtags.get_subclass("AcDbEntity")
    except const.DXFIndexError:
        return
    layer = acdb_entity.get_first_value(8, "0")
    insert = Insert.new(
        dxfattribs={"name": block_name, "layer": layer, "insert": location}
    )
    insert.export_dxf(exporter.tagwriter())


# Planned features: explode complex newer entity types into DXF primitives.
# currently skipped entity types:
# - ACAD_TABLE: graphic as geometry block is available
# --------------------------------------------------------------------------------------
# - all ACIS based entities: tessellated meshes could be exported, but very much work
#   and beyond my current knowledge
# - IMAGE and UNDERLAY: no support possible
# - XRAY and XLINE: no support possible (infinite lines)

# Possible name tags to translate:
# 1 The primary text value for an entity - never a name
# 2 A name: Attribute tag, Block name, and so on. Also used to identify a DXF section or
#   table name
# 3 Other textual or name values - only in DIMENSION a name
# 4 Other textual values - never a name!
# 5 Entity handle expressed as a hexadecimal string (fixed)
# 6 Line type name (fixed)
# 7 Text style name (fixed)
# 8 Layer name (fixed)
# 1001: AppID
# 1003: layer name in XDATA (fixed)
NAME_TAG_CODES = {2, 3, 6, 7, 8, 1001, 1003}


class R12TagWriter(TagWriter):
    def __init__(self, stream: TextIO):
        super().__init__(stream, dxfversion=const.DXF12, write_handles=False)
        self.skip_xdata = False
        self.current_entity = ""
        self.translator = R12NameTranslator()

    def set_stream(self, stream: TextIO) -> None:
        self._stream = stream

    def write_tag(self, tag: DXFTag) -> None:
        code, value = tag
        if code == 0:
            self.current_entity = str(value)
        if code > 999 and self.skip_xdata:
            return
        if code in NAME_TAG_CODES:
            self._stream.write(
                TAG_STRING_FORMAT % (code, self.sanitize_name(code, value))
            )
        else:
            self._stream.write(tag.dxfstr())

    def write_tag2(self, code: int, value) -> None:
        if code > 999 and self.skip_xdata:
            return
        if code == 0:
            self.current_entity = str(value)
        if code in NAME_TAG_CODES:
            value = self.sanitize_name(code, value)
        self._stream.write(TAG_STRING_FORMAT % (code, value))

    def sanitize_name(self, code: int, name: str) -> str:
        # sanitize group code 3 + 4
        # LTYPE - <description> has group code - not a table name
        # STYLE - <font> has group code (3) - not a table name
        # STYLE - <bigfont> has group code (4) - not a table name
        # DIMSTYLE - <dimpost> has group code e.g. "<> mm" (3) - not a table name
        # DIMSTYLE - <dimapost> has group code (4) - not a table name
        # ATTDEF - <prompt> has group code (3) - not a table name
        # DIMENSION - <dimstyle> has group code (3) - is a table name!
        if code == 3 and self.current_entity != "DIMENSION":
            return name
        return self.translator.translate(name)


class SpecialStyleTable:
    def __init__(self, styles: TextstyleTable, extra_styles: TextstyleTable):
        self.styles = styles
        self.extra_styles = extra_styles

    def get_text_styles(self) -> list[Textstyle]:
        entries = list(self.styles.entries.values())
        for name, extra_style in self.extra_styles.entries.items():
            if not self.styles.has_entry(name):
                entries.append(extra_style)
        return entries

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        text_styles = self.get_text_styles()
        tagwriter.write_tag2(0, "TABLE")
        tagwriter.write_tag2(2, "STYLE")
        tagwriter.write_tag2(70, len(text_styles))
        for style in text_styles:
            style.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDTAB")


EOF_STR = "0\nEOF\n"


def detect_max_block_number(names: list[str]) -> int:
    max_number = 0
    for name in names:
        name = name.upper()
        if not name.startswith("*"):
            continue
        try:  # *U10
            number = int(name[2:])
        except ValueError:
            continue
        max_number = max(max_number, number)
    return max_number + 1


class R12Exporter:
    def __init__(self, doc: Drawing, max_sagitta: float = 0.01):
        assert isinstance(doc, Drawing)
        self._doc = doc
        self._tagwriter = R12TagWriter(StringIO())
        self.max_sagitta = float(max_sagitta)  # flattening SPLINE, ELLIPSE
        self.min_spline_segments: int = 4  # flattening SPLINE
        self.min_ellipse_segments: int = 8  # flattening ELLIPSE
        self._extra_doc = ezdxf.new("R12")
        self._next_block_number = detect_max_block_number(
            [br.dxf.name for br in doc.block_records]
        )
        # Exporters are required to convert newer entity types into DXF R12 types.
        # All newer entity types without an exporter will be ignored.
        self.exporters: dict[str, Callable[[R12Exporter, DXFEntity], None]] = {
            "LWPOLYLINE": export_lwpolyline,
            "MESH": export_mesh,
            "SPLINE": export_spline,
            "ELLIPSE": export_ellipse,
            "MTEXT": export_mtext,
            "LEADER": export_virtual_entities,
            "MLEADER": export_virtual_entities,
            "MULTILEADER": export_virtual_entities,
            "MLINE": export_virtual_entities,
            "HATCH": export_hatch,
            "MPOLYGON": export_mpolygon,
            "ACAD_TABLE": export_acad_table,
        }

    def disable_exporter(self, entity_type: str):
        del self.exporters[entity_type]

    @property
    def doc(self) -> Drawing:
        return self._doc

    def tagwriter(self, stream: Optional[TextIO] = None) -> R12TagWriter:
        if stream is not None:
            self._tagwriter.set_stream(stream)
        return self._tagwriter

    def write(self, stream: TextIO) -> None:
        """Write DXF document to text stream."""
        stream.write(self.to_string())

    def to_string(self) -> str:
        """Export DXF document as string."""
        # export layouts before blocks: may create new anonymous blocks
        entities = self.export_layouts_to_string()
        # export blocks before HEADER and TABLES sections: may create new text styles
        blocks = self.export_blocks_to_string()

        return "".join(
            (
                self.export_header_to_string(),
                self.export_tables_to_string(),
                blocks,
                entities,
                EOF_STR,
            )
        )

    def next_block_name(self, char: str) -> str:
        name = f"*{char}{self._next_block_number}"
        self._next_block_number += 1
        return name

    def new_block(self, entity: DXFEntity) -> BlockLayout:
        name = self.next_block_name("U")
        return self._extra_doc.blocks.new(
            name,
            dxfattribs={
                "layer": entity.dxf.get("layer", "0"),
                "flags": const.BLK_ANONYMOUS,
            },
        )

    def export_header_to_string(self) -> str:
        in_memory_stream = StringIO()
        self.doc.header.export_dxf(self.tagwriter(in_memory_stream))
        return in_memory_stream.getvalue()

    def export_tables_to_string(self) -> str:
        # DXF R12 does not support XDATA in tables according Autodesk DWG TrueView
        in_memory_stream = StringIO()
        tables = self.doc.tables
        preserve_table = tables.styles
        tables.styles = SpecialStyleTable(self.doc.styles, self._extra_doc.styles)  # type: ignore

        tagwriter = self.tagwriter(in_memory_stream)
        tagwriter.skip_xdata = True
        tables.export_dxf(tagwriter)
        tables.styles = preserve_table
        tagwriter.skip_xdata = False
        return in_memory_stream.getvalue()

    def export_blocks_to_string(self) -> str:
        in_memory_stream = StringIO()
        self._tagwriter.set_stream(in_memory_stream)

        self._write_section_header("BLOCKS")
        for block_record in self.doc.block_records:
            if block_record.is_any_paperspace and not block_record.is_active_paperspace:
                continue
            name = block_record.dxf.name.lower()
            if name in ("$model_space", "$paper_space"):
                # These block names collide with the translated names of the *Model_Space
                # and the *Paper_Space blocks.
                continue
            self._export_block_record(block_record)

        extra_blocks = self.get_extra_blocks()
        while len(extra_blocks):
            for block_record in extra_blocks:
                self._export_block_record(block_record)
                self.discard_extra_block(block_record.dxf.name)
            # block record export can create further blocks
            extra_blocks = self.get_extra_blocks()

        self._write_endsec()
        return in_memory_stream.getvalue()

    def discard_extra_block(self, name: str) -> None:
        self._extra_doc.block_records.discard(name)

    def get_extra_blocks(self) -> list[BlockRecord]:
        return [
            br for br in self._extra_doc.block_records if br.dxf.name.startswith("*U")
        ]

    def explode_mtext(self, mtext: MText, layout: GenericLayoutType):
        with MTextExplode(layout, self._extra_doc) as xpl:
            xpl.explode(mtext, destroy=False)

    def export_layouts_to_string(self) -> str:
        in_memory_stream = StringIO()
        self._tagwriter.set_stream(in_memory_stream)

        self._write_section_header("ENTITIES")
        self.export_entity_space(self.doc.modelspace().entity_space)
        self.export_entity_space(self.doc.paperspace().entity_space)
        self._write_endsec()
        return in_memory_stream.getvalue()

    def _export_block_record(self, block_record: BlockRecord):
        tagwriter = self._tagwriter
        assert block_record.block is not None
        block_record.block.export_dxf(tagwriter)
        if not block_record.is_any_layout:
            self.export_entity_space(block_record.entity_space)
        assert block_record.endblk is not None
        block_record.endblk.export_dxf(tagwriter)

    def export_entity_space(self, space: EntitySpace):
        tagwriter = self._tagwriter
        for entity in space:
            if entity.MIN_DXF_VERSION_FOR_EXPORT > const.DXF12 or isinstance(
                entity, DXFTagStorage
            ):
                exporter = self.exporters.get(entity.dxftype())
                if exporter:
                    exporter(self, entity)
                    continue
                if entity.proxy_graphic:
                    export_proxy_graphic(self, entity)
            else:
                entity.export_dxf(tagwriter)

    def _write_section_header(self, name: str) -> None:
        self._tagwriter.write_str(f"  0\nSECTION\n  2\n{name}\n")

    def _write_endsec(self) -> None:
        self._tagwriter.write_tag2(0, "ENDSEC")



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.math import Vec2
from ezdxf.addons.drawing import layout


class TestPage:
    def test_ISO_A0(self):
        page = layout.Page(*layout.PAGE_SIZES["ISO A0"])
        assert page.width == 1189
        assert page.height == 841
        assert page.units == layout.Units.mm
        assert page.width_in_mm == 1189
        assert page.height_in_mm == 841

    def test_ANSI_A(self):
        page = layout.Page(*layout.PAGE_SIZES["ANSI A"])
        assert page.width == 11
        assert page.height == 8.5
        assert page.units == layout.Units.inch
        assert page.width_in_mm == 279.4
        assert page.height_in_mm == 215.9

    def test_screen_size_in_pixels(self):
        page = layout.Page(800, 600, layout.Units.px)
        assert page.width_in_mm == 211.7
        assert page.height_in_mm == 158.8


class TestDetectFinalPage:
    """The page size detected automatically if Page.width or Page.height is 0."""

    def test_page_size_scale_1(self):
        """1 DXF drawing unit are represented by 1mm in the output drawing"""
        content_size = Vec2(100, 200)
        page = layout.final_page_size(
            content_size, layout.Page(0, 0, layout.Units.mm), layout.Settings(scale=1)
        )
        assert page.width == 100
        assert page.height == 200

    def test_page_size_scale_50(self):
        """50 DXF drawing unit are represented by 1mm in the output drawing"""
        content_size = Vec2(1000, 2000)
        page = layout.final_page_size(
            content_size,
            layout.Page(0, 0, layout.Units.mm),
            layout.Settings(scale=1 / 50),
        )
        assert page.width == 20
        assert page.height == 40

    def test_page_size_includes_margins_sc1(self):
        content_size = Vec2(100, 200)
        page = layout.final_page_size(
            content_size,
            layout.Page(0, 0, layout.Units.mm, layout.Margins.all2(20, 50)),
            layout.Settings(),
        )
        assert page.width == 100 + 50 + 50
        assert page.height == 200 + 20 + 20

    def test_page_size_includes_margins_sc50(self):
        content_size = Vec2(1000, 2000)
        page = layout.final_page_size(
            content_size,
            layout.Page(0, 0, layout.Units.mm, layout.Margins.all2(20, 50)),
            layout.Settings(scale=1 / 50),
        )
        assert page.width == 20 + 50 + 50
        assert page.height == 40 + 20 + 20

    def test_page_size_limited_page_height(self):
        content_size = Vec2(1000, 2000)
        page = layout.final_page_size(
            content_size,
            layout.Page(0, 0, layout.Units.mm, layout.Margins.all(0), max_height=841),
            layout.Settings(scale=1),
        )
        assert page.height == 841
        assert page.width == pytest.approx(420.5)

    def test_page_size_limited_page_width(self):
        content_size = Vec2(2000, 1000)
        page = layout.final_page_size(
            content_size,
            layout.Page(0, 0, layout.Units.mm, layout.Margins.all(0), max_width=841),
            layout.Settings(scale=1),
        )
        assert page.height == pytest.approx(420.5)
        assert page.width == 841


class TestFitToPage:
    @pytest.fixture(scope="class")
    def page(self):
        return layout.Page(200, 100)

    @pytest.fixture(scope="class")
    def page_with_margins(self):
        return layout.Page(220, 120, margins=layout.Margins.all(10))

    def test_stretch_width(self, page):
        factor = layout.fit_to_page(Vec2(100, 10), page)
        assert factor == pytest.approx(2)

    def test_stretch_height(self, page):
        factor = layout.fit_to_page(Vec2(10, 20), page)
        assert factor == pytest.approx(5)

    def test_shrink_width(self, page):
        factor = layout.fit_to_page(Vec2(400, 10), page)
        assert factor == pytest.approx(0.5)

    def test_shrink_height(self, page):
        factor = layout.fit_to_page(Vec2(50, 200), page)
        assert factor == pytest.approx(0.5)

    def test_stretch_width_margins(self, page_with_margins):
        factor = layout.fit_to_page(Vec2(100, 10), page_with_margins)
        assert factor == pytest.approx(2)

    def test_stretch_height_margins(self, page_with_margins):
        factor = layout.fit_to_page(Vec2(10, 20), page_with_margins)
        assert factor == pytest.approx(5)

    def test_shrink_width_margins(self, page_with_margins):
        factor = layout.fit_to_page(Vec2(400, 10), page_with_margins)
        assert factor == pytest.approx(0.5)

    def test_shrink_height_margins(self, page_with_margins):
        factor = layout.fit_to_page(Vec2(50, 200), page_with_margins)
        assert factor == pytest.approx(0.5)


class TestSettings:
    def test_rotate_content(self):
        settings = layout.Settings(content_rotation=90)
        assert settings.content_rotation == 90

    def test_invalid_rotate_content_angle_raises_exception(self):
        with pytest.raises(ValueError):
            _ = layout.Settings(content_rotation=45)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
# created 2019-03-06
import math

import pytest
from ezdxf.entities.mtext import MText
from ezdxf.layouts import VirtualLayout
from ezdxf.lldxf import const
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.colors import rgb2int

MTEXT = """0
MTEXT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbMText
 10
0
 20
0
 30
0
40
1.0
71
1
1

73
1
"""


@pytest.fixture
def entity():
    e = MText.from_text(MTEXT)
    return e


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "MTEXT" in ENTITY_CLASSES


def test_default_init():
    entity = MText()
    assert entity.dxftype() == "MTEXT"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_virtual_text_attribute():
    """The MText content is stored in multiple tags (1, 3, 3, ...) and cannot
    be supported as a simple DXF tag.
    """
    entity = MText()
    assert entity.is_supported_dxf_attrib("text") is True

    entity.dxf.text = "Hello"
    assert entity.text == "Hello"
    assert entity.dxf.text == "Hello"


def test_setup_by_virtual_text_attribute():
    entity = MText.new(dxfattribs={"text": "content"})
    assert entity.text == "content"
    assert entity.dxf.text == "content"


def test_hasattr_for_virtual_text_attribute_is_always_false():
    entity = MText()
    entity.dxf.text = "Hello"
    assert entity.dxf.hasattr("text") is False


def test_default_new():
    entity = MText.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "insert": (1, 2, 3),
            "char_height": 1.8,
            "width": 20,
            "defined_height": 30,
            "attachment_point": 3,
            "flow_direction": 3,
            "style": "OpenSans",
            "extrusion": (4, 5, 6),
            "text_direction": (7, 8, 9),
            "rect_width": 42,
            "rect_height": 43,
            "rotation": 50,
            "line_spacing_style": 2,
            "line_spacing_factor": 1.7,
            "box_fill_scale": 1.1,
            "bg_fill": 3,
            "bg_fill_color": 14,
            "bg_fill_true_color": 111222,
            "bg_fill_color_name": "magenta",
            "bg_fill_transparency": 1,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.insert == (1, 2, 3)
    assert entity.dxf.char_height == 1.8
    assert entity.dxf.width == 20
    assert entity.dxf.defined_height == 30
    assert entity.dxf.attachment_point == 3
    assert entity.dxf.flow_direction == 3
    assert entity.dxf.extrusion == (4, 5, 6)
    assert entity.dxf.text_direction == (7, 8, 9)
    assert entity.dxf.rect_width == 42
    assert entity.dxf.rect_height == 43
    assert entity.dxf.rotation == 50
    assert entity.dxf.line_spacing_style == 2
    assert entity.dxf.line_spacing_factor == 1.7
    assert entity.dxf.box_fill_scale == 1.1
    assert entity.dxf.bg_fill == 3
    assert entity.dxf.bg_fill_color == 14
    assert entity.dxf.bg_fill_true_color == 111222
    assert entity.dxf.bg_fill_color_name == "magenta"
    assert entity.dxf.bg_fill_transparency == 1


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.insert == (0, 0, 0)
    assert entity.dxf.char_height == 1.0
    assert entity.dxf.attachment_point == 1
    assert entity.dxf.line_spacing_style == 1
    assert entity.text == ""


def test_write_dxf():
    entity = MText.from_text(MTEXT)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(MTEXT)
    assert result == expected


def test_expected_python_backslash_decoding():
    s = (
        r"Swiss 721 (helvetica-like)\P\P\pt7.1875,8.38542;{\H0.6667x;Regular"
        r"\P\P\pl0.899488,t8.38542;Swiss Light^I\fSwis721 Lt BT|b0|i0|c0|p34"
        r";\C5;abcdefghijABCDEFGHIJ123456789!@#$%^ &*()\P\Ftxt.shx|c1;\P\C25"
        r"6;Swiss Light Italic\Ftxt.shx|c0;^I\fSwis721 Lt BT|b0|i"
    )
    assert len(s) == 253


@pytest.fixture
def layout():
    return VirtualLayout()


@pytest.mark.parametrize(
    "text",
    [
        "test\ntext",  # single new line
        "test\r\ntext",  # single new line
        "test\ntext\rtext",  # ignore a single '\r'
    ],
)
def test_required_escaping_of_line_endings_at_dxf_export(layout, text):
    txt = layout.add_mtext(text)
    collector = TagCollector(optional=True)
    txt.export_dxf(collector)
    for tag in collector.tags:
        if tag[0] == 1:
            assert tag[1].count(r"\P") == 1
            assert "\n" not in tag[1]
            assert "\r" not in tag[1]


@pytest.mark.parametrize(
    "text",
    [
        "a new mtext",
        "0123456789" * 25 + "a new mtext",
        "0123456789" * 25 + "abcdefghij" * 25,
    ],
)
def test_new_long_mtext(layout, text):
    mtext = layout.add_mtext(text)
    assert text == mtext.text
    assert text == mtext.plain_text()


def test_default_text_chunk_size_of_250_chars(layout):
    text = "0123456789" * 25 + "abcdefghij" * 25 + "a new mtext"
    mtext = layout.add_mtext(text)
    collector = TagCollector()
    mtext.export_dxf(collector)
    first, second, last = [tag for tag in collector.tags if tag[0] in (1, 3)]

    assert first.code == 3
    assert len(first.value) == 250
    assert text.startswith(first.value)

    assert second.code == 3
    assert len(second.value) == 250
    assert text[250:500] == second.value

    assert last.code == 1
    assert text.endswith(last.value)


def test_text_direction_wins_over_rotation(layout):
    mtext = layout.add_mtext("TEST")
    mtext.dxf.text_direction = (1, 1, 0)  # 45 deg
    mtext.dxf.rotation = 30
    assert mtext.get_rotation() == 45, (
        "Text direction should have higher priority than text rotation,"
        "if both are present."
    )


def test_set_rotation_replaces_text_direction(layout):
    mtext = layout.add_mtext("TEST")
    mtext.dxf.text_direction = (1, 1, 0)  # 45 deg
    mtext.set_rotation(30)
    assert mtext.get_rotation() == 30
    assert mtext.dxf.hasattr("text_direction") is False


def test_append_text(layout):
    mtext = layout.add_mtext("abc")
    mtext += "def"
    assert mtext.text == "abcdef"


def test_set_location(layout):
    mtext = layout.add_mtext("TEST")
    mtext.set_location(
        insert=(3, 4),
        rotation=15,
        attachment_point=const.MTEXT_MIDDLE_CENTER,
    )
    assert const.MTEXT_MIDDLE_CENTER == mtext.dxf.attachment_point
    assert mtext.dxf.rotation == 15
    assert (3, 4, 0) == mtext.dxf.insert


def test_set_bg_color(layout):
    mtext = layout.add_mtext("TEST").set_bg_color(2)
    assert mtext.dxf.bg_fill == 1
    assert mtext.dxf.bg_fill_color == 2
    assert (
        mtext.dxf.hasattr("box_fill_scale") is True
    ), "box_fill_scale attribute must exists, else AutoCAD complains"


def test_set_bg_true_color(layout):
    mtext = layout.add_mtext("TEST").set_bg_color((10, 20, 30), scale=2)
    assert mtext.dxf.bg_fill == 1
    assert mtext.dxf.bg_fill_true_color == rgb2int((10, 20, 30))
    assert mtext.dxf.box_fill_scale == 2
    assert (
        mtext.dxf.hasattr("bg_fill_color") is True
    ), "bg_fill_color attribute must exists, else AutoCAD complains"


def test_delete_bg_color(layout):
    mtext = layout.add_mtext("TEST").set_bg_color(None)
    # AutoCAD always complains about anything if bg_fill is set to 0,
    # so I delete all related tags.
    assert mtext.dxf.hasattr("bg_fill") is False
    assert mtext.dxf.hasattr("bg_fill_color") is False
    assert mtext.dxf.hasattr("bg_fill_true_color") is False
    assert mtext.dxf.hasattr("bg_fill_color_name") is False
    assert mtext.dxf.hasattr("box_fill_scale") is False


def test_set_bg_canvas_color(layout):
    mtext = layout.add_mtext("TEST").set_bg_color("canvas")
    assert mtext.dxf.bg_fill == 3
    assert (
        mtext.has_dxf_attrib("bg_fill_color") is True
    ), "bg_fill_color attribute must exists, else AutoCAD complains"
    assert (
        mtext.has_dxf_attrib("box_fill_scale") is True
    ), "box_fill_scale attribute must exists, else AutoCAD complains"


def test_set_text_frame_only(layout):
    # special case, text frame only with scaling factor = 1.5
    mtext = layout.add_mtext("TEST").set_bg_color(None, text_frame=True)
    assert mtext.dxf.bg_fill == const.MTEXT_TEXT_FRAME
    assert mtext.dxf.hasattr("bg_fill_color") is False
    assert mtext.dxf.hasattr("bg_fill_true_color") is False
    assert mtext.dxf.hasattr("bg_fill_color_name") is False
    assert mtext.dxf.hasattr("box_fill_scale") is False


def test_transform_interface():
    mtext = MText()
    mtext.dxf.insert = (1, 0, 0)
    mtext.translate(1, 2, 3)
    assert mtext.dxf.insert == (2, 2, 3)


def test_bg_fill_flags():
    mtext = MText.new()
    mtext.dxf.bg_fill = 0  # bg fill off
    mtext.dxf.bg_fill = 1  # bg fill as color
    assert mtext.dxf.bg_fill == 1
    mtext.dxf.bg_fill = 2  # bg fill as window color
    assert mtext.dxf.bg_fill == 2
    mtext.dxf.bg_fill = 3  # bg fill as background color
    assert mtext.dxf.bg_fill == 3
    mtext.dxf.bg_fill = 0x10  # text frame?
    assert mtext.dxf.bg_fill == 0x10
    mtext.dxf.bg_fill = 4  # invalid flag
    assert mtext.dxf.bg_fill == 0
    mtext.dxf.bg_fill = 0x20  # invalid flag
    assert mtext.dxf.bg_fill == 0


def test_get_text_direction_from_text_direction():
    mtext = MText()
    mtext.dxf.text_direction = (1, 2, 3)
    assert mtext.get_text_direction().isclose((1, 2, 3))


def test_get_text_direction_from_rotation():
    mtext = MText()
    mtext.dxf.rotation = 45
    s = math.sin(math.radians(mtext.dxf.rotation))
    assert mtext.get_text_direction().isclose((s, s, 0))


def test_convert_rotation_to_text_direction():
    mtext = MText()
    mtext.dxf.rotation = 45
    mtext.convert_rotation_to_text_direction()
    assert mtext.dxf.hasattr("rotation") is False
    assert mtext.dxf.hasattr("text_direction") is True
    old = mtext.dxf.text_direction

    # calling again, does nothing
    mtext.convert_rotation_to_text_direction()
    assert mtext.dxf.hasattr("rotation") is False
    assert mtext.dxf.hasattr("text_direction") is True
    assert old.isclose(mtext.dxf.text_direction)


def test_get_ucs():
    mtext = MText()
    mtext.dxf.insert = (1, 2, 3)
    m = mtext.ucs().matrix
    assert m.origin.isclose((1, 2, 3))
    assert m.ux.isclose((1, 0, 0))
    assert m.uy.isclose((0, 1, 0))
    assert m.uz.isclose((0, 0, 1))



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import math
from ezdxf.math import basic_transformation, BoundingBox, Matrix44
from ezdxf.path import shapes, bbox, triangulate


def test_unit_circle():
    circle = shapes.unit_circle()
    assert circle.has_curves
    assert len(circle) == 4, "expected 4 cubic Bèzier segments"
    assert circle.start.isclose((1, 0, 0))
    assert circle.end.isclose((1, 0, 0))


def test_scale_circle():
    m = shapes.elliptic_transformation(radius=2)
    circle = shapes.unit_circle(transform=m)
    assert circle.start.isclose((2, 0, 0))


def test_move_circle():
    m = shapes.elliptic_transformation(center=(-1, 0, 0))
    circle = shapes.unit_circle(transform=m)
    assert circle.start.isclose((0, 0, 0))


def test_wedge():
    wedge = shapes.wedge(0, math.pi / 2)
    assert wedge.has_curves
    assert wedge.has_lines
    assert wedge.is_closed is True
    assert len(wedge) == 3, "expected 2 lines and 1 cubic Bèzier segment"
    assert wedge.start.isclose((0, 0, 0)), "has to start at the center"
    assert wedge.end.isclose((0, 0, 0)), "has to end at the center"
    extends = bbox([wedge])
    assert extends.extmin.isclose((0.0, 0.0))
    assert extends.extmax.isclose((1.0, 1.0))


def test_ellipse():
    m = shapes.elliptic_transformation(ratio=0.5)
    circle = shapes.unit_circle(transform=m)
    extends = bbox([circle])
    assert extends.extmin.isclose((-1.0, -0.5))
    assert extends.extmax.isclose((1.0, 0.5))


def test_rotated_ellipse():
    m = shapes.elliptic_transformation(ratio=0.5, rotation=math.pi / 2)
    circle = shapes.unit_circle(transform=m)
    extends = bbox([circle])
    assert extends.extmin.isclose((-0.5, -1.0))
    assert extends.extmax.isclose((0.5, 1.0))


@pytest.mark.parametrize(
    "w,h",
    [
        (0, 0),
        (1, 0),
        (0, 1),
        (-1, 1),
        (1, -1),
    ],
)
def test_invalid_rect_dimensions(w, h):
    with pytest.raises(ValueError):
        shapes.rect(w, h)


def test_rectangle():
    rect = shapes.rect(3, 2)
    extends = bbox([rect])
    assert extends.extmin.isclose((-1.5, -1.0))
    assert extends.extmax.isclose((1.5, 1.0))
    assert rect.is_closed is True


def test_transformed_rectangle():
    m = basic_transformation(z_rotation=math.pi / 2)
    rect = shapes.rect(3, 2, transform=m)
    extends = bbox([rect])
    assert extends.extmin.isclose((-1.0, -1.5))
    assert extends.extmax.isclose((1.0, 1.5))


def test_ngon():
    # basic ngon tool is tested here: test_702_render_forms.py
    square = shapes.ngon(4, length=1.0)
    # first vertex is on th x-axis => square is rotated about 45 deg:
    extends = bbox([square])
    d = math.sin(math.pi / 4)
    assert extends.extmin.isclose((-d, -d))
    assert extends.extmax.isclose((d, d))
    assert square.is_closed is True


def test_star():
    # basic star tool is tested here: test_702_render_forms.py
    star4 = shapes.star(4, r1=1, r2=0.5)
    # first vertex is on th x-axis!
    extends = bbox([star4])

    assert extends.extmin.isclose((-1, -1))
    assert extends.extmax.isclose((1, 1))
    assert star4.is_closed is True


def test_gear():
    # basic gear tool is tested here: test_702_render_forms.py
    gear = shapes.gear(16, 0.1, 0.3, 0.2, 1.0)
    # first vertex is on th x-axis!
    extends = bbox([gear])
    assert extends.extmin.isclose((-0.999, -0.999), abs_tol=1e-3)
    assert extends.extmax.isclose((0.999, 0.999), abs_tol=1e-3)
    assert gear.is_closed is True


@pytest.mark.parametrize("ccw", [True, False])
def test_helix_positive_pitch_goes_up(ccw):
    h = shapes.helix(radius=2.0, pitch=1.0, turns=10, ccw=ccw)
    bbox = BoundingBox(h.flattening(0.1))
    assert bbox.extmax.isclose((2, 2, 10))
    assert bbox.extmin.isclose((-2, -2, 0))
    assert h.end.isclose((2, 0, 10))


@pytest.mark.parametrize("ccw", [True, False])
def test_helix_negative_pitch_goes_down(ccw):
    h = shapes.helix(radius=2.0, pitch=-1.0, turns=10, ccw=ccw)
    bbox = BoundingBox(h.flattening(0.1))
    assert bbox.extmax.isclose((2, 2, 0))
    assert bbox.extmin.isclose((-2, -2, -10))
    assert h.end.isclose((2, 0, -10))


def test_triangulate_doughnut():
    circle0 = shapes.unit_circle(segments=8)
    circle1 = circle0.transform(Matrix44.scale(3))
    triangles = list(triangulate([circle1, circle0], 0.01))
    assert len(triangles) == 258


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.document import CREATED_BY_EZDXF, WRITTEN_BY_EZDXF


@pytest.fixture(params=["R12", "R2000"])
def doc(request):
    return ezdxf.new(request.param)


def test_created_by_ezdxf_metadata(doc):
    metadata = doc.ezdxf_metadata()
    assert CREATED_BY_EZDXF in metadata
    assert metadata[CREATED_BY_EZDXF].startswith(ezdxf.__version__)


def test_written_by_ezdxf_metadata(doc, tmp_path):
    doc.saveas(tmp_path / "ez.dxf")
    metadata = doc.ezdxf_metadata()
    assert WRITTEN_BY_EZDXF in metadata
    assert metadata[WRITTEN_BY_EZDXF].startswith(ezdxf.__version__)


def test_add_custom_metadata(doc):
    custom = "CUSTOM"
    metadata = doc.ezdxf_metadata()
    metadata[custom] = custom
    assert custom in metadata
    assert metadata[custom] == custom


def test_non_existing_key_raises_key_error(doc):
    metadata = doc.ezdxf_metadata()
    with pytest.raises(KeyError):
        _ = metadata["XY_UNKNOWN"]


def test_modify_metadata(doc):
    modified = "MODIFIED"
    metadata = doc.ezdxf_metadata()
    metadata[CREATED_BY_EZDXF] = modified
    assert metadata[CREATED_BY_EZDXF] == modified


def test_delete_metadata(doc):
    metadata = doc.ezdxf_metadata()
    del metadata[CREATED_BY_EZDXF]
    assert metadata.get(CREATED_BY_EZDXF) == ""
    assert CREATED_BY_EZDXF not in metadata


def test_delete_non_existing_metadata_raises_KeyError(doc):
    metadata = doc.ezdxf_metadata()
    with pytest.raises(KeyError):
        del metadata["XY_UNKNOWN"]


def test_discard_non_existing_metadata_without_exception(doc):
    metadata = doc.ezdxf_metadata()
    metadata.discard("XY_UNKNOWN")
    assert "XY_UNKNOWN" not in metadata


def test_write_and_read_metadata(doc, tmp_path):
    custom = "CUSTOM"
    metadata = doc.ezdxf_metadata()
    del metadata[CREATED_BY_EZDXF]
    metadata[custom] = custom
    doc.saveas(tmp_path / "ez.dxf")

    doc2 = ezdxf.readfile(tmp_path / "ez.dxf")
    metadata2 = doc2.ezdxf_metadata()
    assert metadata2[WRITTEN_BY_EZDXF].startswith(ezdxf.__version__)
    assert CREATED_BY_EZDXF not in metadata2, "should be deleted"
    assert metadata2[custom] == custom, "expected custom metadata"


def test_key_and_data_is_limited_to_255_chars(doc):
    custom = "CUSTOM" * 60
    metadata = doc.ezdxf_metadata()
    metadata[custom] = custom
    assert custom in metadata
    # max string length for DXF R12 without line endings
    custom255 = custom[:255]
    assert metadata[custom255] == custom255


def test_escape_line_endings_in_key_and_data(doc):
    custom = "CUSTOM\n"
    metadata = doc.ezdxf_metadata()
    metadata[custom] = custom
    assert custom in metadata
    assert (
        metadata[custom] == "CUSTOM\\P"
    ), "line ending should be replaced by \\P"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2023, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.entities.spatial_filter import SpatialFilter


def test_setup_from_tags():
    sp_filter = SpatialFilter.from_text(SPATIAL_FILTER_0)
    assert len(sp_filter.boundary_vertices) == 2
    assert sp_filter.dxf.extrusion == (0, 0, 1)
    assert sp_filter.dxf.origin == (0, 0, 0)
    assert sp_filter.dxf.is_clipping_enabled == 1
    assert sp_filter.dxf.has_front_clipping_plane == 0
    assert sp_filter.dxf.front_clipping_plane_distance == 0.0
    assert sp_filter.dxf.has_back_clipping_plane == 0
    assert sp_filter.dxf.back_clipping_plane_distance == 0.0

def test_loaded_transform_matrix():
    sp_filter = SpatialFilter.from_text(SPATIAL_FILTER_0)
    m = sp_filter.transform_matrix
    assert m.get_row(0) == (1, 0, 0, 0)
    assert m.get_row(1) == (0, 1, 0, 0)
    assert m.get_row(2) == (0, 0, 1, 0)
    assert m.get_row(3) == (0, 0, 0, 1)

def test_loaded_inverse_insert_matrix():
    sp_filter = SpatialFilter.from_text(SPATIAL_FILTER_0)
    m = sp_filter.inverse_insert_matrix
    assert m.get_row(0) == (1, 0, 0, 0)
    assert m.get_row(1) == (0, 1, 0, 0)
    assert m.get_row(2) == (0, 0, 1, 0)
    assert m.get_row(3) == (-240.0, -165.0, 0, 1)

def test_dxf_export():
    sp_filter = SpatialFilter.from_text(SPATIAL_FILTER_0)
    tag_collector = TagCollector()
    sp_filter.export_dxf(tag_collector)
    expected_tags = basic_tags_from_text(SPATIAL_FILTER_0)
    for tag, expected in zip(tag_collector.tags, expected_tags):
        assert tag == expected


SPATIAL_FILTER_0 = """  0
SPATIAL_FILTER
  5
A7
102
{ACAD_REACTORS
330
A6
102
}
330
A6
100
AcDbFilter
100
AcDbSpatialFilter
 70
  2
 10
220.2
 20
152.1
 10
270.1
 20
206.8
210
0.0
220
0.0
230
1.0
 11
0.0
 21
0.0
 31
0.0
 71
  1
 72
  0
 73
  0
 40
1.0
 40
0.0
 40
0.0
 40
-240.0
 40
0.0
 40
1.0
 40
0.0
 40
-165.0
 40
0.0
 40
0.0
 40
1.0
 40
0.0
 40
1.0
 40
0.0
 40
0.0
 40
0.0
 40
0.0
 40
1.0
 40
0.0
 40
0.0
 40
0.0
 40
0.0
 40
1.0
 40
0.0
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Created: 16.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.appid import AppID


@pytest.fixture
def appid():
    return AppID.new(
        "FFFF",
        dxfattribs={
            "name": "EZDXF",
        },
    )


def test_name(appid):
    assert appid.dxf.name == "EZDXF"



# Copyright (c) 2018-2024 Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.lldxf.packedtags import TagArray, VertexArray
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.lldxf.tagwriter import TagCollector
from ezdxf.math import UCS, Matrix44, Vec3


@pytest.fixture()
def numbers():
    return [1, 2, 3, 4]


def test_tag_array_init(numbers):
    array = TagArray(data=numbers)
    for index, value in enumerate(array.values):
        assert value == numbers[index]


def test_tag_array_clone(numbers):
    array = TagArray(data=numbers)
    array2 = array.clone()
    array2.values[-1] = 9999
    assert array.values[:-1] == array2.values[:-1]
    assert array.values[-1] != array2.values[-1]


def test_inherited_array(numbers):
    class FloatArray(TagArray):
        DTYPE = "f"

    floats = FloatArray(data=numbers)
    for index, value in enumerate(floats.values):
        assert value == numbers[index]


def test_vertex_array_basics():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    assert len(vertices) == 7
    points = list(vertices)
    assert len(points) == 7
    assert tuple(vertices[0]) == (0.0, 0.0, 0.0)
    assert tuple(vertices[1]) == (10.0, 10.0, 10.0)
    # test negative index
    assert tuple(vertices[-1]) == (60.0, 60.0, 60.0)
    with pytest.raises(IndexError):
        _ = vertices[-8]
    with pytest.raises(IndexError):
        _ = vertices[8]


def test_vertex_array_advanced():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    # append()
    vertices.append((70, 70, 70))
    assert len(vertices) == 8
    assert tuple(vertices[-1]) == (70.0, 70.0, 70.0)

    # set vertex
    vertices[0] = (7, 6, 5)
    assert tuple(vertices[0]) == (7, 6, 5)
    assert len(vertices) == 8

    # clear()
    vertices.clear()
    assert len(vertices) == 0

    # extend()
    vertices.extend([(0, 0, 0), (1, 2, 3), (4, 5, 6)])
    assert len(vertices) == 3
    assert tuple(vertices[0]) == (0, 0, 0)
    assert tuple(vertices[1]) == (1, 2, 3)
    assert tuple(vertices[2]) == (4, 5, 6)


def test_vertex_array_delete():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    assert len(vertices) == 7
    assert tuple(vertices[0]) == (0, 0, 0)
    del vertices[0]
    assert tuple(vertices[0]) == (10, 10, 10)
    assert len(vertices) == 6

    del vertices[1]  # (20, 20, 20)
    assert tuple(vertices[1]) == (30, 30, 30)
    assert len(vertices) == 5


def test_vertex_array_delete_slices():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    del vertices[:2]
    assert len(vertices) == 5
    assert tuple(vertices[0]) == (20, 20, 20)

    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    del vertices[::2]
    assert len(vertices) == 3
    assert tuple(vertices[0]) == (10, 10, 10)
    assert tuple(vertices[1]) == (30, 30, 30)
    assert tuple(vertices[2]) == (50, 50, 50)


def test_vertex_array_insert():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    assert tuple(vertices[0]) == (0, 0, 0)
    assert tuple(vertices[1]) == (10, 10, 10)
    vertices.insert(1, (-1, -2, -3))
    assert tuple(vertices[0]) == (0, 0, 0)
    assert tuple(vertices[1]) == (-1, -2, -3)
    assert tuple(vertices[2]) == (10, 10, 10)
    assert len(vertices) == 8


def test_vertex_array_to_dxf_tags():
    tags = ExtendedTags.from_text(SPLINE)
    vertices = VertexArray.from_tags(tags.get_subclass("AcDbSpline"))
    tags = TagCollector.dxftags(vertices)
    assert len(tags) == 7 * 3
    assert tags[0] == (10, 0.0)
    assert tags[3] == (10, 10.0)
    assert tags[-1] == (30, 60.0)


def test_vertext_array_transform_by_ucs():
    vertices = VertexArray()
    vertices.extend([(0, 0, 0), (1, 0, 0), (1, 1, 0)])
    ucs = UCS(origin=(0, 0, 1))
    vertices.transform(ucs.matrix)
    assert Vec3(0, 0, 1).isclose(vertices[0])
    assert Vec3(1, 0, 1).isclose(vertices[1])
    assert Vec3(1, 1, 1).isclose(vertices[2])


def test_vertext_transform():
    vertices = VertexArray()
    vertices.extend([(0, 0, 0), (1, 0, 0), (1, 1, 0)])
    m = Matrix44.translate(0, 0, 1)
    vertices.transform(m)
    assert Vec3(0, 0, 1).isclose(vertices[0])
    assert Vec3(1, 0, 1).isclose(vertices[1])
    assert Vec3(1, 1, 1).isclose(vertices[2])


ROOTDICT = """0
DICTIONARY
5
C
330
0
100
AcDbDictionary
281
1
3
ACAD_COLOR
350
73
3
ACAD_GROUP
350
D
3
ACAD_LAYOUT
350
1A
3
ACAD_MATERIAL
350
72
3
ACAD_MLEADERSTYLE
350
D7
3
ACAD_MLINESTYLE
350
17
3
ACAD_PLOTSETTINGS
350
19
3
ACAD_PLOTSTYLENAME
350
E
3
ACAD_SCALELIST
350
B6
3
ACAD_TABLESTYLE
350
86
3
ACAD_VISUALSTYLE
350
99
3
ACDB_RECOMPOSE_DATA
350
40F
3
AcDbVariableDictionary
350
66
3
DWGPROPS
350
410
"""

SPLINE = """0
SPLINE
5
697
102
{ACAD_REACTORS
330
6E8
102
}
330
1F
100
AcDbEntity
8
1
370
20
100
AcDbSpline
210
0.0
220
0.0
230
1.0
70
8
71
3
72
11
73
7
74
0
42
0.000000001
43
0.0000000001
40
0.0
40
0.0
40
0.0
40
0.0
40
1.0
40
2.0
40
3.0
40
3.0
40
3.0
40
3.0
40
3.0
10
0.0
20
0.0
30
0.0
10
10.
20
10.
30
10.
10
20.
20
20.
30
20.
10
30.
20
30.
30
30.
10
40.
20
40.
30
40.
10
50.
20
50.
30
50.
10
60.
20
60.
30
60.
"""



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
import math

from ezdxf.explode import virtual_boundary_path_entities
from ezdxf.entities import Spline
from ezdxf.math import Vec3, Matrix44, OCSTransform


class TestVirtualEntitiesFromEdgePath:
    # fixture "all_edge_types_hatch" from conftest.py
    @pytest.fixture
    def paths(self, all_edge_types_hatch):
        return virtual_boundary_path_entities(all_edge_types_hatch)

    @pytest.fixture
    def spline(self, paths) -> Spline:
        return cast(Spline, paths[0][5])

    def test_returns_one_path(self, paths):
        assert len(paths) == 1

    def test_path_has_six_entities(self, paths):
        assert len(paths[0]) == 6

    def test_expected_entity_types(self, paths):
        types = [e.dxftype() for e in paths[0]]
        assert types == ["ARC", "ELLIPSE", "LINE", "ELLIPSE", "ARC", "SPLINE"]

    def test_arc_0_parameters(self, paths):
        # source edge is a clockwise oriented ArcEdge
        arc_0 = paths[0][0]
        assert arc_0.dxf.center.isclose((0, 13))
        assert arc_0.dxf.radius == pytest.approx(3)
        # ARC entity is always counter-clockwise oriented
        assert arc_0.dxf.start_angle == pytest.approx(-90)
        assert arc_0.dxf.end_angle == pytest.approx(90)

    def test_ellipse_1_parameters(self, paths):
        # source edge is a clockwise oriented EllipseEdge
        ellipse_1 = paths[0][1]
        assert ellipse_1.dxf.center.isclose((0, 5))
        assert ellipse_1.dxf.major_axis.isclose((0, 5))
        assert ellipse_1.dxf.ratio == pytest.approx(0.6)
        # EllipseEdge angles in degrees - ELLIPSE entity params in radians!
        # ELLIPSE entity is always counter-clockwise oriented
        assert ellipse_1.dxf.start_param == pytest.approx(math.radians(180))
        assert ellipse_1.dxf.end_param == pytest.approx(math.radians(360))

    def test_line_parameters(self, paths):
        # source edge is a LineEdge
        line = paths[0][2]
        assert line.dxf.start.isclose((0, 0))
        assert line.dxf.end.isclose((10, 0))

    def test_ellipse_3_parameters(self, paths):
        # source edge is a counter-clockwise oriented EllipseEdge
        ellipse_3 = paths[0][3]
        assert ellipse_3.dxf.center.isclose((10, 5))
        assert ellipse_3.dxf.major_axis.isclose((0, -5))
        assert ellipse_3.dxf.ratio == pytest.approx(0.6)
        # EllipseEdge angles in degrees - ELLIPSE entity params in radians!
        # ELLIPSE entity is always counter-clockwise oriented
        assert ellipse_3.dxf.start_param == pytest.approx(0)
        assert ellipse_3.dxf.end_param == pytest.approx(math.radians(180))

    def test_arc_4_parameters(self, paths):
        # source edge is a counter-clockwise oriented ArcEdge
        arc_4 = paths[0][4]
        assert arc_4.dxf.center.isclose((10, 13))
        assert arc_4.dxf.radius == pytest.approx(3)
        # ARC entity is always counter-clockwise oriented
        assert arc_4.dxf.start_angle == pytest.approx(270)
        assert arc_4.dxf.end_angle == pytest.approx(450)

    def test_spline_parameters(self, spline):
        # source edge is a SplineEdge
        assert spline.dxf.degree == 3
        assert len(spline.control_points) == 6
        assert len(spline.knots) == 10
        assert len(spline.weights) == 0

    def test_spline_control_points(self, spline):
        expected_control_points = [
            Vec3(10.0, 16.0),
            Vec3(9.028174684192452, 16.0),
            Vec3(6.824943218065775, 12.14285714285714),
            Vec3(3.175056781934232, 19.85714285714287),
            Vec3(0.9718253158075516, 16.0),
            Vec3(0, 16.0),
        ]
        assert (
            all(
                expected_cp.isclose(cp)
                for cp, expected_cp in zip(
                    spline.control_points, expected_control_points
                )
            )
            is True
        )

    def test_spline_knots(self, spline):
        expected_knot_values = [
            0.0,
            0.0,
            0.0,
            0.0,
            2.91547594742265,
            8.746427842267952,
            11.6619037896906,
            11.6619037896906,
            11.6619037896906,
            11.6619037896906,
        ]
        assert spline.knots == pytest.approx(expected_knot_values)


class TestSpatialTransformation:
    # fixture "all_edge_types_hatch" from conftest.py
    @pytest.fixture(scope="class")
    def matrix(self):
        return Matrix44.y_rotate(math.radians(-90))

    @pytest.fixture
    def paths(self, all_edge_types_hatch, matrix):
        all_edge_types_hatch.transform(matrix)
        return virtual_boundary_path_entities(all_edge_types_hatch)

    @pytest.fixture
    def spline(self, paths) -> Spline:
        return cast(Spline, paths[0][5])

    def test_wcs_line(self, paths, matrix):
        line = paths[0][2]
        assert line.dxf.start.isclose(matrix.transform((0, 0)))
        assert line.dxf.end.isclose(matrix.transform((10, 0)))

    def test_wcs_spline(self, spline, matrix):
        expected_control_points = matrix.transform_vertices(
            [
                Vec3(10.0, 16.0),
                Vec3(9.028174684192452, 16.0),
                Vec3(6.824943218065775, 12.14285714285714),
                Vec3(3.175056781934232, 19.85714285714287),
                Vec3(0.9718253158075516, 16.0),
                Vec3(0, 16.0),
            ]
        )
        assert (
            all(
                expected_cp.isclose(cp)
                for cp, expected_cp in zip(
                    spline.control_points, expected_control_points
                )
            )
            is True
        )

    def test_ocs_clockwise_arc(self, paths, matrix):
        arc = paths[0][0]
        ocs_transform = OCSTransform(Vec3(0, 0, 1), matrix)

        wcs_center = ocs_transform.new_ocs.to_wcs(arc.dxf.center)
        assert wcs_center.isclose(matrix.transform((0, 13)))

        expected_start_angle = ocs_transform.transform_deg_angle(-90)
        assert arc.dxf.start_angle == pytest.approx(expected_start_angle)

        expected_end_angle = ocs_transform.transform_deg_angle(90)
        assert arc.dxf.end_angle == pytest.approx(expected_end_angle)

    def test_ocs_counter_clockwise_arc(self, paths, matrix):
        arc = paths[0][4]
        ocs_transform = OCSTransform(Vec3(0, 0, 1), matrix)

        wcs_center = ocs_transform.new_ocs.to_wcs(arc.dxf.center)
        assert wcs_center.isclose(matrix.transform((10, 13)))

        expected_start_angle = ocs_transform.transform_deg_angle(270)
        assert arc.dxf.start_angle == pytest.approx(expected_start_angle)

        expected_end_angle = ocs_transform.transform_deg_angle(450)
        assert arc.dxf.end_angle == pytest.approx(expected_end_angle)

    def test_wcs_clockwise_ellipse(self, paths, matrix):
        ellipse = paths[0][1]
        assert ellipse.dxf.center.isclose(matrix.transform((0, 5)))
        assert ellipse.dxf.major_axis.isclose(
            matrix.transform_direction((0, 5))
        )
        # ellipse normal vector
        assert ellipse.dxf.extrusion.isclose(
            matrix.transform_direction((0, 0, 1))
        )
        # only the center and the major axis are transformed:
        assert ellipse.dxf.start_param == pytest.approx(math.radians(180))
        assert ellipse.dxf.end_param == pytest.approx(math.radians(360))

    def test_wcs_counter_clockwise_ellipse(self, paths, matrix):
        ellipse = paths[0][3]
        assert ellipse.dxf.center.isclose(matrix.transform((10, 5)))
        assert ellipse.dxf.major_axis.isclose(
            matrix.transform_direction((0, -5))
        )
        # ellipse normal vector
        assert ellipse.dxf.extrusion.isclose(
            matrix.transform_direction((0, 0, 1))
        )
        # only the center and the major axis are transformed:
        assert ellipse.dxf.start_param == pytest.approx(math.radians(0))
        assert ellipse.dxf.end_param == pytest.approx(math.radians(180))


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2020 Manfred Moitzi
# License: MIT License
import pytest
import math

from ezdxf.entities.point import Point
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Matrix44
from ezdxf.explode import explode_entity

TEST_CLASS = Point
TEST_TYPE = "POINT"

ENTITY_R12 = """0
POINT
5
0
8
0
10
1.0
20
2.0
30
3.0
"""

ENTITY_R2000 = """0
POINT
5
0
330
0
100
AcDbEntity
8
0
100
AcDbPoint
10
1.0
20
2.0
30
3.0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "location": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.location == (1, 2, 3)
    assert entity.dxf.location.x == 1, "is not Vec3 compatible"
    assert entity.dxf.location.y == 2, "is not Vec3 compatible"
    assert entity.dxf.location.z == 3, "is not Vec3 compatible"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.location.isclose((1, 2, 3))


MALFORMED_POINT = """0
POINT
5
0
330
0
62
7
8
MALFORMED
102
{MOZMAN
8
APP_DATA_LAYER_IGNORED
6
APP_DATA_LTYPE_IGNORED
62
0
102
}
100
AcDbEntity
10
1.0
20
2.0
30
3.0
100
AcDbPoint
6
DOTTED
100
AcDbInvalidSubclassMarker
"""


def test_load_malformed_entity():
    """Missing AcDbPoint subclass marker."""
    entity = TEST_CLASS.from_text(MALFORMED_POINT)
    assert entity.dxf.layer == "MALFORMED"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "DOTTED"
    assert entity.dxf.location.isclose((1, 2, 3))


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    point = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    point.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    point.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_transform():
    point = Point.new(
        dxfattribs={
            "location": (2, 3, 4),
            "extrusion": (0, 1, 0),
            "thickness": 2,
        }
    )
    # 1. rotation - 2. scaling - 3. translation
    m = Matrix44.chain(Matrix44.scale(2, 3, 1), Matrix44.translate(1, 1, 1))
    point.transform(m)
    assert point.dxf.location == (5, 10, 5)
    assert point.dxf.extrusion == (0, 1, 0)
    assert point.dxf.thickness == 6

    angle = math.pi / 4
    point.transform(Matrix44.z_rotate(math.pi / 4))
    assert point.dxf.extrusion.isclose((-math.cos(angle), math.sin(angle), 0))
    assert math.isclose(point.dxf.thickness, 6)


def test_fast_translation():
    point = Point.new(
        dxfattribs={
            "location": (2, 3, 4),
            "extrusion": (0, 1, 0),
            "thickness": 2,
        }
    )
    point.translate(1, 2, 3)
    assert point.dxf.location == (3, 5, 7)


def test_do_not_explode_point_entity():
    point = Point()
    with pytest.raises(TypeError):
        explode_entity(point)


def test_virtual_sub_entities():
    point = Point()
    entities = list(point.virtual_entities())
    assert len(entities) == 1
    e = entities[0]
    assert e.is_copy is True
    assert e.source_of_copy is point




# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.solid import Solid, Trace, Face3d
from ezdxf.lldxf.const import DXF12, DXF2000
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

TEST_CLASS = Solid
TEST_TYPE = "SOLID"

ENTITY_R12 = """0
SOLID
5
0
8
0
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
12
0.0
22
0.0
32
0.0
13
0.0
23
0.0
33
0.0
"""

ENTITY_R2000 = """0
SOLID
5
0
330
0
100
AcDbEntity
8
0
100
AcDbTrace
10
0.0
20
0.0
30
0.0
11
0.0
21
0.0
31
0.0
12
0.0
22
0.0
32
0.0
13
0.0
23
0.0
33
0.0
"""

ELEVATION = """0
SOLID
5
0
8
0
38
2.0
10
0.0
20
0.0
11
0.0
21
0.0
12
0.0
22
0.0
13
0.0
23
0.0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def entity(request):
    return TEST_CLASS.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = TEST_CLASS()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = TEST_CLASS.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "vtx3": (1, 2, 3),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.linetype == "BYLAYER"
    assert entity.dxf.vtx3 == (1, 2, 3)
    assert entity.dxf.vtx3.x == 1, "is not Vec3 compatible"
    assert entity.dxf.vtx3.y == 2, "is not Vec3 compatible"
    assert entity.dxf.vtx3.z == 3, "is not Vec3 compatible"
    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.vtx3 == (0, 0, 0)


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    solid = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    solid.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    solid.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_trace():
    trace = Trace()
    trace[0] = (1, 2, 3)
    trace[1] = (4, 5, 6)
    trace[2] = (7, 8, 9)
    trace[3] = (5, 1, 0)
    assert trace[0] == (1, 2, 3)
    assert trace[1] == (4, 5, 6)
    assert trace[2] == (7, 8, 9)
    assert trace[3] == (5, 1, 0)
    assert trace.dxf.vtx0 == (1, 2, 3)
    assert trace.dxf.vtx1 == (4, 5, 6)
    assert trace.dxf.vtx2 == (7, 8, 9)
    assert trace.dxf.vtx3 == (5, 1, 0)

    collector = TagCollector(dxfversion=DXF2000)
    trace.export_dxf(collector)
    assert collector.tags[0] == (0, "TRACE")
    assert collector.tags[5] == (100, "AcDbTrace")

    # Elevation tag should not be written by default
    assert any(tag[0] == 38 for tag in collector.tags) is False


def test_3dface():
    face = Face3d()
    face.dxf.invisible_edges = 2 + 8
    assert face.is_invisible_edge(0) is False
    assert face.is_invisible_edge(1) is True
    assert face.is_invisible_edge(2) is False
    assert face.is_invisible_edge(3) is True

    assert face.get_edges_visibility() == [True, False, True, False]

    face.dxf.vtx3 = (1, 2, 3)
    assert face.get_edges_visibility() == [True, False, True, False]


    face.dxf.invisible_edges = 0
    face.set_edge_visibility(3, False)
    assert face.dxf.invisible_edges == 8
    face.set_edge_visibility(1, False)
    assert face.dxf.invisible_edges == 10
    face.set_edge_visibility(3, True)
    assert face.dxf.invisible_edges == 2

    collector = TagCollector(dxfversion=DXF2000)
    face.export_dxf(collector)
    assert collector.tags[0] == (0, "3DFACE")
    assert collector.tags[5] == (100, "AcDbFace")


def test_solid_translate():
    solid = Solid()
    solid.dxf.vtx1 = (3, 3, 0)
    solid.translate(1, 1, 0)
    assert solid.dxf.vtx1 == (4, 4, 0)


def test_trace_translate():
    face = Face3d()
    face.dxf.vtx1 = (3, 3, 0)
    face.translate(1, 1, 0)
    assert face.dxf.vtx1 == (4, 4, 0)


def test_solid_reorder_quad_ocs_vertices():
    solid = Solid()
    for index, vertex in enumerate([(0, 0), (1, 0), (0, 1), (1, 1)]):
        solid[index] = vertex

    # reorder weird vertex order:
    assert solid.vertices() == [(0, 0), (1, 0), (1, 1), (0, 1)]


def test_solid_triangle_ocs_vertices():
    solid = Solid()
    for index, vertex in enumerate([(0, 0), (1, 0), (0, 1), (0, 1)]):
        solid[index] = vertex
    assert solid.vertices() == [(0, 0), (1, 0), (0, 1)]


def test_solid_close_triangle_ocs_vertices():
    solid = Solid()
    for index, vertex in enumerate([(0, 0), (1, 0), (0, 1), (0, 1)]):
        solid[index] = vertex
    assert solid.vertices(close=True) == [(0, 0), (1, 0), (0, 1), (0, 0)]


def test_solid_close_quad_ocs_vertices():
    solid = Solid()
    for index, vertex in enumerate([(0, 0), (1, 0), (1, 1), (0, 1)]):
        solid[index] = vertex
    assert solid.vertices(close=True) == [
        (0, 0),
        (1, 0),
        (0, 1),
        (1, 1),
        (0, 0),
    ]


def test_solid_wcs_vertices():
    solid = Solid()
    for index, vertex in enumerate([(0, 0), (1, 0), (0, 1), (1, 1)]):
        solid[index] = vertex
    # reorder weird vertex order:
    assert solid.wcs_vertices() == [(0, 0), (1, 0), (1, 1), (0, 1)]


def test_3dface_quad_vertices():
    face = Face3d()
    for index, vertex in enumerate([(0, 0), (1, 0), (1, 1), (0, 1)]):
        face[index] = vertex
    assert face.get_edges_visibility() == [True, True, True, True]
    # no weird vertex order:
    assert face.wcs_vertices() == [(0, 0), (1, 0), (1, 1), (0, 1)]
    assert face.wcs_vertices(close=True) == [
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 1),
        (0, 0),
    ]


def test_3dface_triangle_vertices():
    face = Face3d()
    for index, vertex in enumerate([(0, 0), (1, 0), (1, 1), (1, 1)]):
        face[index] = vertex
    assert face.get_edges_visibility() == [True, True, True, True]
    # changed in v0.17.2: do not return duplicated vertices at the end
    assert face.wcs_vertices() == [(0, 0), (1, 0), (1, 1)]
    assert face.wcs_vertices(close=True) == [
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 0),
    ]


def test_elevation_group_code_support():
    solid = Solid.from_text(ELEVATION)
    # elevation data is copied to z-axis of vertices:
    assert solid.dxf.hasattr("elevation") is False
    vertices = solid.vertices()
    assert vertices[0] == (0, 0, 2)


def test_do_not_write_elevation_group_code():
    solid = Solid.from_text(ELEVATION)
    collector = TagCollector(dxfversion=DXF12)
    solid.export_dxf(collector)
    # Elevation tag should be written:
    assert any(tag[0] == 38 for tag in collector.tags) is False


MALFORMED_SOLID = """0
SOLID
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbEntity
100
AcDbTrace
10
1.0
20
1.0
30
1.0
11
2.0
21
2.0
31
2.0
100
AcDbTrace
12
3.0
22
3.0
32
3.0
13
4.0
23
4.0
33
4.0
"""


@pytest.mark.parametrize(
    "text,cls",
    [
        (MALFORMED_SOLID, Solid),
        (MALFORMED_SOLID.replace("SOLID", "TRACE"), Trace),
        (MALFORMED_SOLID.replace("SOLID", "3DFACE"), Face3d),
    ],
    ids=["SOLID", "TRACE", "3DFACE"],
)
def test_load_malformed_solid(text, cls):
    entity = cls.from_text(text)
    assert cls.DXFTYPE in text
    assert entity.dxf.layer == "LY_EZDXF"
    assert entity.dxf.linetype == "LT_EZDXF"
    assert entity.dxf.color == 7
    assert entity.dxf.vtx0.isclose((1, 1, 1))
    assert entity.dxf.vtx1.isclose((2, 2, 2))
    assert entity.dxf.vtx2.isclose((3, 3, 3))
    assert entity.dxf.vtx3.isclose((4, 4, 4))



# Copyright (c) 2011-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Iterator, cast, Optional
from itertools import chain
import logging

from ezdxf.lldxf import const
from ezdxf.entities import entity_linker

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFEntity, DXFTagStorage, BlockRecord, DXFGraphic
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.lldxf.tags import Tags


logger = logging.getLogger("ezdxf")


class StoredSection:
    def __init__(self, entities: list[Tags]):
        self.entities = entities

    def export_dxf(self, tagwriter: AbstractTagWriter):
        # (0, SECTION) (2, NAME) is stored in entities
        for entity in self.entities:
            tagwriter.write_tags(entity)
        # ENDSEC not stored in entities !!!
        tagwriter.write_str("  0\nENDSEC\n")


class EntitySection:
    """:class:`EntitiesSection` is just a proxy for :class:`Modelspace` and
    active :class:`Paperspace` linked together.
    """

    def __init__(
        self,
        doc: Optional[Drawing] = None,
        entities: Optional[Iterable[DXFEntity]] = None,
    ):
        self.doc = doc
        if entities is not None:
            self._build(iter(entities))

    def __iter__(self) -> Iterator[DXFEntity]:
        """Returns an iterator for all entities of the modelspace and the active
        paperspace.
        """
        assert self.doc is not None
        layouts = self.doc.layouts
        for entity in chain(layouts.modelspace(), layouts.active_layout()):
            yield entity

    def __len__(self) -> int:
        """Returns the count of all entities in the modelspace and the active paperspace.
        """
        assert self.doc is not None
        layouts = self.doc.layouts
        return len(layouts.modelspace()) + len(layouts.active_layout())

    # none public interface

    def _build(self, entities: Iterator[DXFEntity]) -> None:
        assert self.doc is not None
        section_head = cast("DXFTagStorage", next(entities))
        if section_head.dxftype() != "SECTION" or section_head.base_class[
            1
        ] != (2, "ENTITIES"):
            raise const.DXFStructureError(
                "Critical structure error in ENTITIES section."
            )

        def add(entity: DXFGraphic):
            handle = entity.dxf.owner
            # higher priority for owner handle
            paperspace = 0
            if handle == msp_layout_key:
                paperspace = 0
            elif handle == psp_layout_key:
                paperspace = 1
            elif entity.dxf.hasattr(
                "paperspace"
            ):  # paperspace flag as fallback
                paperspace = entity.dxf.paperspace

            if paperspace:
                psp.add_entity(entity)
            else:
                msp.add_entity(entity)

        msp = cast("BlockRecord", self.doc.block_records.get("*Model_Space"))
        psp = cast("BlockRecord", self.doc.block_records.get("*Paper_Space"))
        msp_layout_key: str = msp.dxf.handle
        psp_layout_key: str = psp.dxf.handle
        linked_entities = entity_linker()
        # Don't store linked entities (VERTEX, ATTRIB, SEQEND) in entity space
        for entity in entities:
            # No check for valid entities here:
            # Use the audit- or the recover module to fix invalid DXF files!
            if not linked_entities(entity):
                add(entity)  # type: ignore

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        assert self.doc is not None
        layouts = self.doc.layouts
        tagwriter.write_str("  0\nSECTION\n  2\nENTITIES\n")
        # Just write *Model_Space and the active *Paper_Space into the
        # ENTITIES section.
        layouts.modelspace().entity_space.export_dxf(tagwriter)
        layouts.active_layout().entity_space.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
import os
import ezdxf
from ezdxf.lldxf.const import versions_supported_by_new

NONE_ASCII = "äöüÄÖÜß±ØáàÀÁóòÓÒéèÉÈ"


@pytest.fixture(params=versions_supported_by_new)
def doc(request):
    return ezdxf.new(request.param)


def test_write_and_read_unicode(doc, tmpdir):
    msp = doc.modelspace()
    msp.add_text(NONE_ASCII)
    filename = str(tmpdir.join("none_ascii_%s.dxf" % doc.dxfversion))
    try:
        doc.saveas(filename)
    except ezdxf.DXFError as e:
        pytest.fail(
            "DXFError: {0} for DXF version {1}".format(str(e), doc.dxfversion)
        )
    assert os.path.exists(filename)

    doc = ezdxf.readfile(filename)
    text = doc.modelspace().query("TEXT")
    assert len(text) == 1
    assert text[0].dxf.text == NONE_ASCII



# Copyright (c) 2020-2021, Matthew Broadway
# License: MIT License
from __future__ import annotations

from ezdxf.addons.drawing.backend import BackendInterface
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.math import Vec3


def draw_rect(points: list[Vec3], color: Color, out: BackendInterface):
    from ezdxf.addons.drawing.properties import BackendProperties

    props = BackendProperties(color=color)
    for a, b in zip(points, points[1:]):
        out.draw_line(a, b, props)



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
# Test only basic features of Cython implementation,
# Full testing and compatibility check with Python implementation
# is located in test suite 603.

import pytest
import math

cyvec = pytest.importorskip("ezdxf.acc.vector")
Vec2 = cyvec.Vec2


def test_default_constructor():
    v = Vec2()
    assert v.x == 0
    assert v.y == 0


def test_is_immutable():
    v = Vec2()
    with pytest.raises(AttributeError):
        v.x = 1
    with pytest.raises(AttributeError):
        v.y = 1


def test_init_x_y():
    v = Vec2(1, 2)
    assert v.x == 1
    assert v.y == 2


def test_init_2_tuple():
    v = Vec2((1, 2))
    assert v.x == 1
    assert v.y == 2


def test_init_vec2():
    v = Vec2(Vec2(1, 2))
    assert v.x == 1
    assert v.y == 2


def test_init_3_tuple():
    v = Vec2((1, 2, 3))
    assert v.x == 1
    assert v.y == 2


def test_init_vec3():
    v = Vec2(cyvec.Vec3(1, 2, 3))
    assert v.x == 1
    assert v.y == 2


def test_init_type_error():
    with pytest.raises(TypeError):
        Vec2(1)
    with pytest.raises(TypeError):
        Vec2(1, 2, 3, 4)
    with pytest.raises(TypeError):
        Vec2("xyz")
    with pytest.raises(TypeError):
        Vec2("xyz", "abc")


def test_compare():
    assert Vec2(1, 2) == Vec2(1, 2)
    assert Vec2(1, 2) == (1, 2)
    assert (1, 2) == Vec2(1, 2)
    assert Vec2(1, 2) != (
        1,
        2.0000000000001,
    ), "__eq__() should use the full floating point precision"
    assert Vec2(1, 2) != (2, 1)


def test_lower_than():
    assert Vec2(1, 2) < Vec2(2, 2)
    assert Vec2(2, 2) > Vec2(1, 2)


def test_is_null():
    assert Vec2().is_null is True


def test_bool():
    assert bool(Vec2()) is False


def test_does_not_support_slicing():
    with pytest.raises(TypeError):
        _ = Vec2(2, 1)[:]


@pytest.mark.parametrize(
    "angle", [90, -90, 600, -600, 0.0, 360.0, -360.0, 720.0, -720.0]
)
def test_normalize_deg_angle(angle):
    assert cyvec._normalize_deg_angle(angle) == angle % 360.0


@pytest.mark.parametrize(
    "angle",
    [
        3.0,
        -3.0,
        10.0,
        -10.0,
        0.0,
        math.tau,
        -math.tau,
        2 * math.tau,
        -2 * math.tau,
    ],
)
def test_normalize_rad_angle(angle):
    assert cyvec._normalize_rad_angle(angle) == angle % math.tau


if __name__ == "__main__":
    pytest.main([__file__])
